#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
"""Functions for plotting artis estimators and internal structure.

Examples are temperatures, populations, heating/cooling rates.
"""

import argparse
import math
import string
import typing as t
from collections.abc import Collection
from collections.abc import Sequence
from pathlib import Path

import argcomplete
import matplotlib.axes as mplax
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import polars as pl
from matplotlib import ticker

import artistools as at

colors_tab10: list[str] = list(plt.get_cmap("tab10")(np.linspace(0, 1.0, 10)))

# reserve colours for these elements
elementcolors = {"Fe": colors_tab10[0], "Ni": colors_tab10[1], "Co": colors_tab10[2]}

VARIABLE_ALIASES = {"T_e": "Te", "n_e": "nne", "T_R": "TR", "T_J": "TJ"}


def get_elemcolor(atomic_number: int | None = None, elsymbol: str | None = None) -> str | npt.NDArray[t.Any]:
    """Get the colour of an element from the reserved color list (reserving a new one if needed)."""
    assert (atomic_number is None) != (elsymbol is None)
    if atomic_number is not None:
        elsymbol = at.get_elsymbol(atomic_number)
    assert elsymbol is not None
    # assign a new colour to this element if needed

    return elementcolors.setdefault(elsymbol, colors_tab10[len(elementcolors)])


def get_ylabel(variable: str) -> str:
    return at.estimators.get_variablelongunits(variable) or at.estimators.get_units_string(variable)


def adjust_lightness(color: t.Any, amount: float = 0.5) -> tuple[float, float, float]:
    import colorsys

    import matplotlib.colors as mc

    try:
        c = mc.cnames[color]
    except (SyntaxWarning, KeyError, TypeError):
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def plot_data(
    dfplotdata: pl.DataFrame | pl.LazyFrame,
    ax: mplax.Axes,
    label: str | None,
    args: argparse.Namespace,
    startfromzero: bool = False,
    **plotkwargs: t.Any,
) -> None:
    dfplotdata = dfplotdata.lazy()

    # Calculate the average line and optionally, the min-max bounding area
    dflinepoints = (
        dfplotdata.group_by("xvalue_binned", maintain_order=True)
        .agg(
            yvalue_binned=(pl.col("yvalue") * pl.col("celltsweight")).sum() / pl.col("celltsweight").sum(),
            yvalue_binned_min=pl.col("yvalue").min(),
            yvalue_binned_max=pl.col("yvalue").max(),
        )
        .sort("xvalue_binned")
        .drop_nans()
    )

    filterfunc = at.get_filterfunc(args)
    if filterfunc is not None:
        dflinepoints = dflinepoints.with_columns(
            pl.col("yvalue_binned").map_batches(filterfunc, return_dtype=pl.self_dtype())
        )

    if startfromzero:
        dflinepoints = pl.concat([
            pl.LazyFrame(
                {
                    "xvalue_binned": [0.0],
                    **{
                        col: [dflinepoints.select(pl.col(col).head(1)).collect().item()]
                        for col in dflinepoints.collect_schema().names()
                        if col != "xvalue_binned"
                    },
                },
                schema=dflinepoints.collect_schema(),
            ),
            dflinepoints,
        ]).lazy()

    xvalues_binned = dflinepoints.select(["xvalue_binned"]).collect().get_column("xvalue_binned")
    yvalues_binned = dflinepoints.select(["yvalue_binned"]).collect().get_column("yvalue_binned")

    (plotobj,) = ax.plot(xvalues_binned, yvalues_binned, label=label, **plotkwargs)
    color = plotobj.get_color()

    if args.markers:
        plotkwargs_markers = plotkwargs | {
            "linestyle": "None",
            "marker": ".",
            "markersize": 5,
            "color": adjust_lightness(color, 1.5),
            # "alpha": 0.4,
            "markeredgewidth": 0,
            "zorder": -1,
        }
        plotkwargs_markers.pop("dashes", None)
        plotkwargs_markers.pop("label", None)
        if dfplotdata.select(pl.len() > 10000).collect().item():
            plotkwargs_markers["rasterized"] = True
        # plot the markers first
        ax.plot(
            dfplotdata.select("xvalue").collect().to_series(),
            dfplotdata.select("yvalue").collect().to_series(),
            **plotkwargs_markers,
        )

    else:
        yvalues_binned_min = dflinepoints.select("yvalue_binned_min").collect().get_column("yvalue_binned_min")
        yvalues_binned_max = dflinepoints.select("yvalue_binned_max").collect().get_column("yvalue_binned_max")
        plotobj = ax.fill_between(
            xvalues_binned, yvalues_binned_min, yvalues_binned_max, alpha=0.2, color=color, linewidth=0, zorder=-2
        )


def plot_init_abundances(
    ax: mplax.Axes,
    specieslist: list[str],
    estimators: pl.LazyFrame,
    seriestype: str,
    startfromzero: bool,
    args: argparse.Namespace,
    **plotkwargs: t.Any,
) -> None:
    if seriestype == "initmasses":
        estimators = estimators.with_columns(
            (pl.col(massfraccol) * pl.col("mass_g") / 1.989e33).alias(
                f"init_mass_{massfraccol.removeprefix('init_X_')}"
            )
            for massfraccol in estimators.collect_schema().names()
            if massfraccol.startswith("init_X_")
        )
        ax.set_ylabel(r"Initial mass per x point [M$_\odot$]")
        valuetype = "init_mass_"
    else:
        assert seriestype == "initabundances"
        ax.set_ylim(1e-20, 1.0)
        ax.set_ylabel("Initial mass fraction")
        valuetype = "init_X_"

    for speciesstr in specieslist:
        splitvariablename = speciesstr.split("_")
        elsymbol = splitvariablename[0].strip(string.digits)
        atomic_number = at.get_atomic_number(elsymbol)

        linestyle = "-"
        if speciesstr.lower() in {"ni_56", "ni56", "56ni"}:
            expr_yvalue = pl.col(f"{valuetype}Ni56")
            linelabel = "$^{56}$Ni"
            linestyle = "--"
        elif speciesstr.lower() in {"ni_stb", "ni_stable"}:
            expr_yvalue = pl.col(f"{valuetype}{elsymbol}") - pl.col(f"{valuetype}Ni56")
            linelabel = "Stable Ni"
        elif speciesstr.lower() in {"co_56", "co56", "56co"}:
            expr_yvalue = pl.col(f"{valuetype}Co56")
            linelabel = "$^{56}$Co"
        elif speciesstr.lower() in {"fegrp", "ffegroup"}:
            expr_yvalue = pl.col(f"{valuetype}Fegroup")
            linelabel = "Fe group"
        else:
            linelabel = speciesstr
            expr_yvalue = pl.col(f"{valuetype}{elsymbol}")

        plotkwargs["color"] = get_elemcolor(atomic_number=atomic_number)
        plotkwargs.setdefault("linewidth", 1.5)
        series = estimators.with_columns(celltsweight=pl.col("rho") * pl.col("deltavol_deltat"), yvalue=expr_yvalue)

        if "linestyle" not in plotkwargs:
            plotkwargs["linestyle"] = linestyle

        plot_data(series, ax=ax, args=args, startfromzero=startfromzero, label=linelabel, **plotkwargs)


def plot_average_ionisation(
    ax: mplax.Axes,
    params: Sequence[str],
    estimators: pl.LazyFrame,
    startfromzero: bool,
    args: argparse.Namespace | None = None,
    **plotkwargs: t.Any,
) -> None:
    if args is None:
        args = argparse.Namespace()

    ax.set_ylabel("Average ion charge")

    for paramvalue in params:
        print(f"  plotting averageionisation {paramvalue}")
        atomic_number = at.get_atomic_number(paramvalue)

        color = get_elemcolor(atomic_number=atomic_number)
        elsymb = at.get_elsymbol(atomic_number)
        if f"nnelement_{elsymb}" not in estimators.collect_schema().names():
            msg = f"ERROR: No element data found for {paramvalue}"
            raise ValueError(msg)

        ioncols = [col for col in estimators.collect_schema().names() if col.startswith(f"nnion_{elsymb}_")]
        ioncharges = [at.decode_roman_numeral(col.removeprefix(f"nnion_{elsymb}_")) - 1 for col in ioncols]
        ax.set_ylim(0.0, max(ioncharges) + 0.1)
        expr_charge_per_nuc = pl.sum_horizontal([
            ioncharge * pl.col(ioncol) for ioncol, ioncharge in zip(ioncols, ioncharges, strict=True)
        ]) / pl.col(f"nnelement_{elsymb}")

        dfplotdata = estimators.with_columns(
            celltsweight=pl.col(f"nnelement_{elsymb}") * pl.col("deltavol_deltat"), yvalue=expr_charge_per_nuc
        ).filter(pl.col(f"nnelement_{elsymb}") > 0.0)

        plot_data(
            dfplotdata=dfplotdata,
            ax=ax,
            args=args,
            startfromzero=startfromzero,
            label=paramvalue,
            color=color,
            **plotkwargs,
        )


def plot_levelpop(
    ax: mplax.Axes,
    xlist: Sequence[int | float] | npt.NDArray[np.floating],
    seriestype: str,
    params: Sequence[str],
    timestepslist: Sequence[int],
    mgilist: Sequence[int | Sequence[int]],
    modelpath: str | Path,
    startfromzero: bool,
    args: argparse.Namespace,
    **plotkwargs: t.Any,
) -> None:
    if seriestype == "levelpopulation_dn_on_dvel":
        ax.set_ylabel("dN/dV [{}km$^{{-1}}$ s]")
        ax.yaxis.set_major_formatter(at.plottools.ExponentLabelFormatter(ax.get_ylabel()))
    elif seriestype == "levelpopulation":
        ax.set_ylabel("X$_{{i}}$ [{}/cm³]")
        ax.yaxis.set_major_formatter(at.plottools.ExponentLabelFormatter(ax.get_ylabel()))
    else:
        raise ValueError

    modeldata = (
        at.inputmodel.get_modeldata(modelpath, derived_cols=["mass_g", "volume"])[0]
        .collect()
        .to_pandas(use_pyarrow_extension_array=True)
    )

    adata = (
        at.atomic.get_levels(modelpath)
        .with_columns(
            levels=pl.col("levels").map_elements(
                lambda x: x.to_pandas(use_pyarrow_extension_array=True), return_dtype=pl.Object
            )
        )
        .to_pandas(use_pyarrow_extension_array=True)
    )

    arr_tdelta = at.get_timestep_times(modelpath, loc="delta")
    for paramvalue in params:
        paramsplit = paramvalue.split(" ")
        atomic_number = at.get_atomic_number(paramsplit[0])
        ion_stage = at.decode_roman_numeral(paramsplit[1])
        levelindex = int(paramsplit[2])

        ionlevels = adata.query("Z == @atomic_number and ion_stage == @ion_stage").iloc[0].levels
        levelname = ionlevels.iloc[levelindex].levelname
        label = (
            f"{at.get_ionstring(atomic_number, ion_stage, style='chargelatex')} level {levelindex}:"
            f" {at.nltepops.texifyconfiguration(levelname)}"
        )

        print(f"plot_levelpop {label}")

        # level index query goes outside for caching granularity reasons
        dfnltepops = at.nltepops.read_files(
            modelpath, dfquery=f"Z=={atomic_number:.0f} and ion_stage=={ion_stage:.0f}"
        ).query("level==@levelindex")

        ylist = []
        for modelgridindex in mgilist:
            valuesum = 0.0
            tdeltasum = 0.0
            # print(f'modelgridindex {modelgridindex} timesteps {timesteps}')

            for timestep in timestepslist:
                levelpop = (
                    dfnltepops.query(
                        "modelgridindex==@modelgridindex and timestep==@timestep and Z==@atomic_number"
                        " and ion_stage==@ion_stage and level==@levelindex"
                    )
                    .iloc[0]
                    .n_NLTE
                )

                valuesum += levelpop * arr_tdelta[timestep]
                tdeltasum += arr_tdelta[timestep]

            if seriestype == "levelpopulation_dn_on_dvel":
                assert isinstance(modelgridindex, int)
                deltav = modeldata.loc[modelgridindex].vel_r_max_kmps - modeldata.loc[modelgridindex].vel_r_min_kmps
                ylist.append(valuesum / tdeltasum * modeldata.loc[modelgridindex].volume / deltav)
            else:
                ylist.append(valuesum / tdeltasum)

        plot_data(
            pl.DataFrame({"xvalue": xlist, "yvalue": ylist}),
            ax=ax,
            args=args,
            startfromzero=startfromzero,
            label=label,
            **plotkwargs,
        )


def get_iontuple(ionstr: str) -> tuple[int, str | int]:
    """Decode into atomic number and parameter, e.g., [(26, 1), (26, 2), (26, 'ALL'), (26, 'Fe56')]."""
    if ionstr in at.get_elsymbolslist():
        return (at.get_atomic_number(ionstr), "ALL")

    # a space separates the element symbol from the ionstage, e.g. Fe II
    if " " in ionstr:
        return (at.get_atomic_number(ionstr.split(" ", maxsplit=1)[0]), at.decode_roman_numeral(ionstr.split(" ")[1]))

    # for element symbol with a mass number after it, e.g. Fe56
    if ionstr.rstrip("-0123456789") in at.get_elsymbolslist():
        atomic_number = at.get_atomic_number(ionstr.rstrip("-0123456789"))
        return (atomic_number, ionstr)

    # for element and ionstage without a space, e.g. FeII
    for elsymb in at.get_elsymbolslist():
        if ionstr.startswith(elsymb):
            possible_roman = at.decode_roman_numeral(ionstr.removeprefix(elsymb))
            if possible_roman > 0:
                return (at.get_atomic_number(elsymb), possible_roman)

    atomic_number = at.get_atomic_number(ionstr.split("_", maxsplit=1)[0])
    return (atomic_number, ionstr)


def get_column_name(seriestype: str, atomic_number: int, ion_stage: str | int) -> tuple[str, str]:
    ionstr = at.get_ionstring(atomic_number, ion_stage, sep="_", style="spectral")
    if seriestype == "populations":
        if ion_stage == "ALL":
            elsymbol = at.get_elsymbol(atomic_number)
            return f"nnelement_{elsymbol}", ionstr
        if isinstance(ion_stage, str) and ion_stage.startswith(at.get_elsymbol(atomic_number)):
            # not really an ion_stage but an isotope name
            return f"nniso_{ion_stage}", ionstr
        return f"nnion_{ionstr}", ionstr
    return f"{seriestype}_{ionstr}", ionstr


def plot_multi_ion_series(
    ax: mplax.Axes,
    startfromzero: bool,
    seriestype: str,
    ionlist: Sequence[str],
    estimators: pl.LazyFrame,
    modelpath: str | Path,
    args: argparse.Namespace,
    ymin: float | None = None,
    ymax: float | None = None,
    **plotkwargs: t.Any,
) -> None:
    """Plot an ion-specific property, e.g., populations."""
    # if seriestype == 'populations':
    #     ax.yaxis.set_major_locator(ticker.MultipleLocator(base=0.10))

    plotted_something = False

    iontuplelist = [get_iontuple(ionstr) for ionstr in ionlist]
    iontuplelist.sort()
    print(f"Subplot with ions: {iontuplelist}")

    missingions: set[tuple[int, str | int]] = set()
    try:
        if not args.classicartis:
            compositiondata = at.get_composition_data(modelpath)
            for atomic_number, ion_stage in iontuplelist:
                if (
                    not hasattr(ion_stage, "lower")
                    and not args.classicartis
                    and compositiondata.filter(
                        (pl.col("Z") == atomic_number)
                        & (pl.col("lowermost_ion_stage") <= ion_stage)
                        & (pl.col("uppermost_ion_stage") >= ion_stage)
                    ).is_empty()
                ):
                    missingions.add((atomic_number, ion_stage))

    except FileNotFoundError:
        print("WARNING: Could not read an ARTIS compositiondata.txt file to check ion availability")
        for atomic_number, ion_stage in iontuplelist:
            ionstr = at.get_ionstring(atomic_number, ion_stage, sep="_", style="spectral")
            if f"nnion_{ionstr}" not in estimators.collect_schema().names():
                missingions.add((atomic_number, ion_stage))

    if missingions:
        print(f" Warning: Can't plot {seriestype} for {missingions} because these ions are not in compositiondata.txt")

    iontuplelist = [iontuple for iontuple in iontuplelist if iontuple not in missingions]
    lazyframes = []
    for atomic_number, ion_stage in iontuplelist:
        colname, ionstr = get_column_name(seriestype, atomic_number, ion_stage)
        expr_yvals = pl.col(colname)
        print(f"  plotting {seriestype} {ionstr.replace('_', ' ')}")

        if seriestype != "populations" or args.poptype == "absolute":
            expr_normfactor = pl.lit(1)
        elif args.poptype == "elpop":
            elsymbol = at.get_elsymbol(atomic_number)
            expr_normfactor = pl.col(f"nnelement_{elsymbol}")
        elif args.poptype == "totalpop":
            expr_normfactor = pl.col("nntot")
        elif args.poptype in {"radialdensity", "cylradialdensity"}:
            # get the volumetric number density to later be multiplied by the surface area of a sphere or cylinder
            expr_normfactor = pl.lit(1)
        elif args.poptype == "cumulative":
            # multiply by volume to get number of particles
            expr_normfactor = pl.lit(1) / pl.col("volume")
        else:
            raise AssertionError

        # convert volumetric number density to radial density
        if args.poptype == "radialdensity":
            expr_yvals *= 4 * math.pi * pl.col("vel_r_mid").mean().pow(2)
        elif args.poptype == "cylradialdensity":
            expr_yvals *= 2 * math.pi * pl.col("vel_rcyl_mid").mean()

        if args.poptype == "cumulative":
            expr_yvals = expr_yvals.cum_sum()

        lazyframes.append(
            estimators.with_columns(celltsweight=pl.col("deltavol_deltat"), yvalue=expr_yvals / expr_normfactor)
        )

    for seriesindex, (iontuple, dfseries) in enumerate(zip(iontuplelist, pl.collect_all(lazyframes), strict=True)):
        atomic_number, ion_stage = iontuple
        plotlabel = str(
            ion_stage
            if hasattr(ion_stage, "lower") and ion_stage != "ALL"
            else at.get_ionstring(atomic_number, ion_stage, style="chargelatex")
        )

        color = get_elemcolor(atomic_number=atomic_number)

        # linestyle = ['-.', '-', '--', (0, (4, 1, 1, 1)), ':'] + [(0, x) for x in dashes_list][ion_stage - 1]
        dashes: tuple[float, ...] = ()
        styleindex = 0
        if isinstance(ion_stage, str):
            if ion_stage != "ALL":
                # isotopic abundance
                if args.colorbyion:
                    color = f"C{seriesindex % 10}"
                else:
                    styleindex = seriesindex
        else:
            assert isinstance(ion_stage, int)
            if args.colorbyion:
                color = f"C{(ion_stage - 1) % 10}"
            else:
                styleindex = ion_stage - 1

        dashes_list = [(), (3, 1, 1, 1), (1.5, 1.5), (6, 3), (1, 3)]
        dashes = dashes_list[styleindex % len(dashes_list)]

        linewidth_list = [1.0, 1.0, 1.0, 0.7, 0.7]
        linewidth = linewidth_list[styleindex % len(linewidth_list)] * 1.5

        if plotkwargs.get("linestyle", "solid") != "None":
            plotkwargs["dashes"] = dashes

        plot_data(
            dfseries,
            linewidth=linewidth,
            label=plotlabel,
            ax=ax,
            args=args,
            startfromzero=startfromzero,
            color=color,
            **plotkwargs,
        )
        plotted_something = True

    if seriestype == "populations":
        if args.poptype == "absolute":
            ax.set_ylabel(r"Number density $\left[\rm{cm}^{-3}\right]$")
        elif args.poptype == "elpop":
            # elsym = at.get_elsymbol(atomic_number)
            ax.set_ylabel(r"X$_{i}$/X$_{\rm element}$")
        elif args.poptype == "totalpop":
            ax.set_ylabel(r"X$_{i}$/X$_{\rm tot}$")
        elif args.poptype == "radialdensity":
            ax.set_ylabel(r"Radial density dN/dr $\left[\rm{cm}^{-1}\right]$")
        elif args.poptype == "cylradialdensity":
            ax.set_ylabel(r"Cylindrical radial density dN/drcyl $\left[\rm{cm}^{-1}\right]$")
        elif args.poptype == "cumulative":
            ax.set_ylabel(r"Cumulative particle count")
        else:
            raise AssertionError
    else:
        ax.set_ylabel(at.estimators.get_varname_formatted(seriestype))

    if plotted_something and ax.get_yscale() == "log":
        ymin, ymax = ax.get_ylim()
        ymin = max(ymin, ymax / 1e10)
        ax.set_ylim(bottom=ymin)
        # make space for the legend
        new_ymax = ymax * 10 ** (0.1 * math.log10(ymax / ymin))
        if ymin > 0 and new_ymax > ymin and np.isfinite(new_ymax):
            ax.set_ylim(top=new_ymax)


def plot_series(
    ax: mplax.Axes,
    startfromzero: bool,
    variable: str | pl.Expr,
    showlegend: bool,
    estimators: pl.LazyFrame,
    args: argparse.Namespace,
    nounits: bool = False,
    **plotkwargs: t.Any,
) -> None:
    """Plot something like Te or TR."""
    if isinstance(variable, pl.Expr):
        colexpr = variable
    else:
        assert variable in estimators.collect_schema().names(), f"Variable {variable} not found in estimators"
        colexpr = pl.col(variable)

    variablename = colexpr.meta.output_name()

    serieslabel = at.estimators.get_varname_formatted(variablename)
    units_string = at.estimators.get_units_string(variablename)

    if showlegend:
        linelabel = serieslabel
        if not nounits:
            linelabel += units_string
    else:
        ax.set_ylabel(serieslabel + units_string)
        linelabel = None

    series = estimators.with_columns(celltsweight=pl.col("deltavol_deltat"), yvalue=colexpr)

    if variablename in (dictcolors := {"Te": "red", "heating_gamma": "blue", "cooling_adiabatic": "blue"}):
        plotkwargs.setdefault("color", dictcolors[variablename])
    plotkwargs.setdefault("linewidth", 1.5)

    print(f"  plotting {variablename}")
    plot_data(series, ax=ax, label=linelabel, args=args, startfromzero=startfromzero, **plotkwargs)


def get_xlist(
    xvariable: str, estimators: pl.LazyFrame, timestepslist: Collection[int] | None, args: t.Any
) -> tuple[list[float | int], list[int], list[int], pl.LazyFrame]:
    if timestepslist is not None:
        estimators = estimators.filter(pl.col("timestep").is_in(timestepslist))

    if xvariable in {"cellid", "modelgridindex"}:
        estimators = estimators.with_columns(xvalue=pl.col("modelgridindex"))
    elif xvariable == "timestep":
        estimators = estimators.with_columns(xvalue=pl.col("timestep"))
    elif xvariable == "time":
        estimators = estimators.with_columns(xvalue=pl.col("tmid_days"))
    elif xvariable in {"velocity", "beta"}:
        velcolumn = "vel_r_mid"
        scalefactor = 1e5 if xvariable == "velocity" else 29979245800
        estimators = estimators.with_columns(xvalue=(pl.col(velcolumn) / scalefactor))
    else:
        assert xvariable in estimators.collect_schema().names()
        estimators = estimators.with_columns(xvalue=pl.col(xvariable))

    xmin = estimators.select(pl.col("xvalue").min()).collect().item() if args.xmin is None else args.xmin
    xmax = estimators.select(pl.col("xvalue").max()).collect().item() if args.xmax is None else args.xmax

    if (
        args.xbins is None
        and estimators.select(pl.n_unique("xvalue") * pl.n_unique("timestep") < pl.len()).collect().item()
    ):
        print("There are multiple plot points per x value. Using automatic bins (use -xbins N to change this)")
        args.xbins = -1
        args.colorbyion = True

    if args.xbins is not None and args.xbins < 0:
        xdeltamax = estimators.select(pl.col("xvalue").sort().diff().max()).collect().item()
        args.xbins = int((xmax - xmin) / xdeltamax)
        print(
            f"Setting xbins to {args.xbins} based on data range [{xmin}, {xmax}] and largest x interval of {xdeltamax}"
        )
        if args.xbins <= 3:
            print(f"  would have only {args.xbins} bins. Replacing with 25")
            args.xbins = 25

    if args.xbins is not None:
        xbinedges = np.linspace(xmin, xmax, args.xbins)
        xlower = xbinedges[:-1]
        xupper = xbinedges[1:]
        xmids = (xlower + xupper) / 2
        estimators = (
            estimators.with_columns(
                pl.col("xvalue")
                .cut(breaks=list(xbinedges), labels=[str(x) for x in range(-1, len(xbinedges))])
                .cast(pl.Utf8)
                .cast(pl.Int32)
                .alias("xbinindex")
            )
            .filter(pl.col("xbinindex").is_between(0, len(xmids) - 1, closed="both"))
            .join(pl.LazyFrame({"xvalue_binned": xmids}).with_row_index("xbinindex"), on="xbinindex", how="left")
            .drop("xbinindex")
        )
    else:
        estimators = estimators.with_columns(xvalue_binned=pl.col("xvalue"))

    if args.xmin is not None:
        estimators = estimators.filter(pl.col("xvalue") >= args.xmin)

    if args.xmax is not None:
        estimators = estimators.filter(pl.col("xvalue") <= args.xmax)

    estimators = estimators.sort("xvalue")
    xvalues = estimators.select("xvalue").unique().collect().get_column("xvalue")
    assert len(xvalues) > 0, "No data found for x-axis variable"

    return (
        xvalues.to_list(),
        estimators.select(pl.col("modelgridindex").unique()).collect()["modelgridindex"].to_list(),
        estimators.select(pl.col("timestep").unique()).collect()["timestep"].to_list(),
        estimators,
    )


def plot_subplot(
    ax: mplax.Axes,
    timestepslist: list[int],
    xlist: list[float | int],
    startfromzero: bool,
    plotitems: list[t.Any],
    mgilist: list[int],
    modelpath: str | Path,
    estimators: pl.LazyFrame,
    args: argparse.Namespace,
    **plotkwargs: t.Any,
) -> None:
    """Make plot from ARTIS estimators."""
    # these three lists give the x value, modelgridex, and a list of timesteps (for averaging) for each plot of the plot
    showlegend = False
    legend_kwargs = {}
    seriescount = 0
    ylabel = None
    sameylabel = True
    seriesvars = [var for var in plotitems if isinstance(var, str | pl.Expr)]
    seriescount = len(seriesvars)
    print(f"Subplot: {plotitems}")
    for variable in seriesvars:
        variablename = variable.meta.output_name() if isinstance(variable, pl.Expr) else variable
        if ylabel is None:
            ylabel = get_ylabel(variablename)
        elif ylabel != get_ylabel(variablename):
            sameylabel = False
            break
    ymin, ymax = None, None
    for plotitem in plotitems:
        if isinstance(plotitem, str | pl.Expr):
            continue
        seriestype, params = plotitem
        if seriestype == "_ymin":
            ymin = float(params) if isinstance(params, str) else params
            ax.set_ylim(bottom=ymin)

        elif seriestype == "_ymax":
            ymax = float(params) if isinstance(params, str) else params
            ax.set_ylim(top=ymax)

        elif seriestype == "_yscale":
            ax.set_yscale(params)

    for plotitem in plotitems:
        if isinstance(plotitem, str | pl.Expr):
            variablename = plotitem.meta.output_name() if isinstance(plotitem, pl.Expr) else plotitem
            assert isinstance(variablename, str)
            showlegend = seriescount > 1 or len(variablename) > 35 or not sameylabel
            plot_series(
                ax=ax,
                startfromzero=startfromzero,
                variable=plotitem,
                showlegend=showlegend,
                estimators=estimators,
                args=args,
                nounits=sameylabel,
                **plotkwargs,
            )
            if showlegend and sameylabel and ylabel is not None:
                ax.set_ylabel(ylabel)
        else:  # it's a sequence of values
            seriestype, params = plotitem
            showlegend = True
            if isinstance(params, str) and seriestype.startswith("_"):
                continue

            if seriestype in {"initabundances", "initmasses"}:
                assert isinstance(params, list)
                plot_init_abundances(
                    ax=ax,
                    specieslist=params,
                    estimators=estimators,
                    seriestype=seriestype,
                    startfromzero=startfromzero,
                    args=args,
                    **plotkwargs,
                )

            elif seriestype == "levelpopulation" or seriestype.startswith("levelpopulation_"):
                plot_levelpop(
                    ax,
                    xlist,
                    seriestype,
                    params,
                    timestepslist,
                    mgilist,
                    modelpath,
                    startfromzero=startfromzero,
                    args=args,
                )

            elif seriestype == "averageionisation":
                plot_average_ionisation(ax, params, estimators, startfromzero=startfromzero, args=args, **plotkwargs)

            else:
                seriestype, ionlist = plotitem
                if seriestype.startswith("_"):
                    continue
                ax.set_yscale("log")
                if seriestype == "populations" and len(ionlist) > 2 and ax.get_yscale() == "log":
                    legend_kwargs["ncol"] = 2

                plot_multi_ion_series(
                    ax=ax,
                    startfromzero=startfromzero,
                    seriestype=seriestype,
                    ionlist=ionlist,
                    estimators=estimators,
                    modelpath=modelpath,
                    args=args,
                    ymin=ymin,
                    ymax=ymax,
                    **plotkwargs,
                )

    ax.tick_params(right=True)
    if showlegend and not args.nolegend:
        ax.legend(loc="best", handlelength=2, frameon=False, numpoints=1, **legend_kwargs, markerscale=3)


def make_figure(
    modelpath: Path | str,
    timestepslist: Collection[int] | None,
    estimators: pl.LazyFrame,
    xvariable: str,
    plotlist: list[list[t.Any]],
    args: t.Any,
    **plotkwargs: t.Any,
) -> str:
    modelname = at.get_model_name(modelpath)

    fig, axes = plt.subplots(
        nrows=len(plotlist),
        ncols=1,
        sharex=True,
        figsize=(
            args.figscale * at.get_config()["figwidth"] * args.scalefigwidth,
            args.figscale * at.get_config()["figwidth"] * 0.5 * len(plotlist),
        ),
        layout="constrained",
        # tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0},
    )
    if len(plotlist) == 1:
        axes = np.array([axes])

    assert isinstance(axes, np.ndarray)

    for ax in axes:
        ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
    # ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=5))
    if not args.hidexlabel:
        axes[-1].set_xlabel(
            f"{at.estimators.get_varname_formatted(xvariable)}{at.estimators.get_units_string(xvariable)}"
        )

    xlist, mgilist, timestepslist, estimators = get_xlist(
        xvariable=xvariable, estimators=estimators, timestepslist=timestepslist, args=args
    )

    startfromzero = xvariable.startswith("velocity") or xvariable == "beta"
    xmin = args.xmin if args.xmin is not None else min(xlist)
    xmax = args.xmax if args.xmax is not None else max(xlist)

    for ax, plotitems in zip(axes, plotlist, strict=False):
        if xmin != xmax:
            ax.set_xlim(left=xmin, right=xmax)

        plot_subplot(
            ax=ax,
            timestepslist=timestepslist,
            xlist=xlist,
            plotitems=plotitems,
            mgilist=mgilist,
            modelpath=modelpath,
            estimators=estimators,
            startfromzero=startfromzero,
            args=args,
            **plotkwargs,
        )

    if len(set(mgilist)) == 1 and len(timestepslist) > 1:  # single grid cell versus time plot
        figure_title = f"{modelname}\nCell {mgilist[0]}"

        defaultoutputfile = "plotestimators_cell{modelgridindex:03d}.{format}"
        if Path(args.outputfile).is_dir():
            args.outputfile = str(Path(args.outputfile) / defaultoutputfile)

        outfilename = str(args.outputfile).format(modelgridindex=mgilist[0], format=args.format)

    else:
        if args.multiplot:
            strtimestep = f"ts{timestepslist[0]:02d}"
            strtimedays = f"{at.get_timestep_time(modelpath, timestepslist[0]):.2f}d"
        else:
            timesteps_flat = at.flatten_list(timestepslist)
            timestepmin = min(timesteps_flat)
            timestepmax = max(timesteps_flat)

            strtimestep = (
                f"ts{timestepmin:02d}-ts{timestepmax:02d}" if timestepmax != timestepmin else f"ts{timestepmin:02d}"
            )
            dftimesteps = at.get_timesteps(modelpath)
            timelow_days = (
                dftimesteps.filter(pl.col("timestep") == timestepmin).select(pl.col("tstart_days")).collect().item()
            )
            timehigh_days = (
                dftimesteps.filter(pl.col("timestep") == timestepmax).select(pl.col("tend_days")).collect().item()
            )
            strtimedays = f"{timelow_days:.2f}d-{timehigh_days:.2f}d"

        figure_title = f"{modelname}\nTimestep {strtimestep} ({strtimedays})"
        print("  plotting " + figure_title.replace("\n", " "))

        defaultoutputfile = "plotestimators_{timestep}_{timedays}.{format}"
        if Path(args.outputfile).is_dir():
            args.outputfile = str(Path(args.outputfile) / defaultoutputfile)

        assert isinstance(timestepslist, list)
        outfilename = str(args.outputfile).format(timestep=strtimestep, timedays=strtimedays, format=args.format)

    if not args.notitle:
        axes[0].set_title(figure_title, fontsize=10)

    print(f"open {outfilename}")
    fig.savefig(outfilename, dpi=600)

    if args.show:
        plt.show()
    else:
        plt.close()

    return outfilename


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-modelpath", default=".", help="Paths to ARTIS folder (or virtual path e.g. codecomparison/ddc10/cmfgen)"
    )

    parser.add_argument(
        "-modelgridindex", "-cell", "-mgi", type=int, default=None, help="Modelgridindex for time evolution plot"
    )

    parser.add_argument("-timestep", "-ts", nargs="?", help="Timestep number for internal structure plot")

    parser.add_argument("-timedays", "-time", "-t", nargs="?", help="Time in days to plot for internal structure plot")

    parser.add_argument("-timemin", type=float, help="Lower time in days")

    parser.add_argument("-timemax", type=float, help="Upper time in days")

    parser.add_argument("--multiplot", action="store_true", help="Make multiple plots for timesteps in range")

    parser.add_argument("-x", default=None, help="Horizontal axis variable, e.g. velocity, timestep, or time")

    parser.add_argument("-xmin", type=float, default=None, help="Plot range: minimum x value")

    parser.add_argument("-xmax", type=float, default=None, help="Plot range: maximum x value")

    parser.add_argument(
        "-xbins", type=int, default=None, help="Number of x bins between xmax and xmin (or -1 for automatic bin size)"
    )

    parser.add_argument("--hidexlabel", action="store_true", help="Hide the bottom horizontal axis label")

    parser.add_argument("--markers", action="store_true", help="Plot markers instead of shaded area")

    parser.add_argument("-filtermovingavg", type=int, default=0, help="Smoothing length (1 is same as none)")

    parser.add_argument(
        "-filtersavgol",
        nargs=2,
        help="Savitzky-Golay filter. Specify the window_length and polyorder.e.g. -filtersavgol 5 3",
    )

    parser.add_argument("-format", "-f", default="pdf", choices=["pdf", "png"], help="Set format of output plot files")

    parser.add_argument("--makegif", action="store_true", help="Make a gif with time evolution (requires --multiplot)")

    parser.add_argument("--notitle", action="store_true", help="Suppress the top title from the plot")

    parser.add_argument(
        "-plotlist",
        "-plot",
        "-p",
        nargs="*",
        type=str,
        action="append",
        help="List of plots to generate. Specify estimator names or population types. Examples: -plot Te TR -plot nne -plot SrI 'Sr II'",
    )

    parser.add_argument(
        "-ionpoptype",
        "-poptype",
        dest="poptype",
        default="elpop",
        choices=["absolute", "totalpop", "elpop", "radialdensity", "cylradialdensity", "cumulative"],
        help="Plot absolute ion populations, or ion populations as a fraction of total or element population",
    )

    parser.add_argument("--nolegend", action="store_true", help="Suppress the legend from the plot")

    parser.add_argument(
        "-figscale", type=float, default=1.0, help="Scale factor for plot area. 1.0 is for single-column"
    )

    parser.add_argument("-scalefigwidth", type=float, default=1.0, help="Scale factor for plot width.")

    parser.add_argument("--show", action="store_true", help="Show plot before quitting")

    parser.add_argument(
        "-outputfile",
        "-outputpath",
        "-o",
        action="store",
        dest="outputfile",
        type=Path,
        default=Path(),
        help="Filename for PDF file",
    )

    parser.add_argument(
        "--colorbyion", action="store_true", help="Populations plots colored by ion rather than element"
    )

    parser.add_argument(
        "--classicartis", action="store_true", help="Flag to show using output from classic ARTIS branch"
    )

    parser.add_argument(
        "-readonlymgi",
        default=False,
        choices=["alongaxis", "cone"],  # plan to extend this to e.g. 2D slice
        help="Option to read only selected mgi and choice of which mgi to select. Choose which axis with args.axis",
    )

    parser.add_argument(
        "-axis",
        default="+z",
        choices=["+x", "-x", "+y", "-y", "+z", "-z"],
        help="Choose an axis for use with args.readonlymgi. Hint: for negative use e.g. -axis=-z",
    )


def main(args: argparse.Namespace | None = None, argsraw: Sequence[str] | None = None, **kwargs: t.Any) -> None:
    """Plot ARTIS estimators."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)
        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    modelpath = Path(args.modelpath)

    modelname = at.get_model_name(modelpath)

    should_use_all_timesteps = (
        not args.timedays
        and not args.timestep
        and (args.modelgridindex is not None or args.x in {None, "time", "timestep"})
    )

    if should_use_all_timesteps:
        args.timestep = f"0-{len(at.get_timestep_times(modelpath)) - 1}"
        if args.x is None:
            args.x = "time"
            print(f"Setting x variable to {args.x}")
    elif args.x is None:
        args.x = "velocity"
        print(f"Setting x variable to {args.x}")

    (timestepmin, timestepmax, args.timemin, args.timemax) = at.get_time_range(
        modelpath, args.timestep, args.timemin, args.timemax, args.timedays
    )

    if args.readonlymgi:
        args.sliceaxis = args.axis[1]
        assert args.axis[0] in {"+", "-"}
        args.positive_axis = args.axis[0] == "+"

        axes = ["x", "y", "z"]
        axes.remove(args.sliceaxis)
        args.other_axis1 = axes[0]
        args.other_axis2 = axes[1]

    print(
        f"Plotting estimators for '{modelname}' timesteps {timestepmin} to {timestepmax} "
        f"({args.timemin:.1f} to {args.timemax:.1f}d)"
    )

    if args.readonlymgi:
        if args.readonlymgi == "alongaxis":
            print(f"Getting mgi along {args.axis} axis")
            dfselectedcells = at.inputmodel.slice1dfromconein3dmodel.get_profile_along_axis(args=args)

        elif args.readonlymgi == "cone":
            print(f"Getting mgi lying within a cone around {args.axis} axis")
            dfselectedcells = at.inputmodel.slice1dfromconein3dmodel.make_cone(args)
        else:
            msg = f"Invalid args.readonlymgi: {args.readonlymgi}"
            raise ValueError(msg)
        dfselectedcells = dfselectedcells[dfselectedcells["rho"] > 0]
        args.modelgridindex = list(dfselectedcells["inputcellid"])

    timesteps_included = list(range(timestepmin, timestepmax + 1))
    if args.classicartis:
        import artistools.estimators.estimators_classic

        estimatorsdict = artistools.estimators.estimators_classic.read_classic_estimators(modelpath)
        assert estimatorsdict is not None
        estimators = pl.DataFrame([
            {"timestep": ts, "modelgridindex": mgi, **estimvals} for (ts, mgi), estimvals in estimatorsdict.items()
        ]).lazy()
    else:
        estimators = at.estimators.scan_estimators(
            modelpath=modelpath, modelgridindex=args.modelgridindex, timestep=tuple(timesteps_included)
        )

    estimators, modelmeta = at.estimators.join_cell_modeldata(estimators=estimators, modelpath=modelpath, verbose=False)
    if estimators.select(pl.len()).collect().item() == 0:
        print("No data was found for the requested timesteps/cells.")
        estimators = at.estimators.scan_estimators(modelpath=modelpath)
        print("Cells with data: ")
        print(estimators.select(pl.col("modelgridindex").unique().sort()).collect().to_series().to_list())
        print("Timesteps with data: ")
        print(estimators.select(pl.col("timestep").unique().sort()).collect().to_series().to_list())
        return

    if args.modelgridindex is None:
        estimators = estimators.filter(pl.col("vel_r_mid") <= modelmeta["vmax_cmps"])

    estimators = estimators.with_columns(deltavol_deltat=pl.col("volume") * pl.col("twidth_days"))

    plotlist = args.plotlist or [
        # [["initabundances", ["Fe", "Ni_stable", "Ni_56"]]],
        # ['heating_dep', 'heating_coll', 'heating_bf', 'heating_ff',
        #  ['_yscale', 'linear']],
        # ['cooling_adiabatic', 'cooling_coll', 'cooling_fb', 'cooling_ff',
        #  ['_yscale', 'linear']],
        # [
        #     (pl.col("heating_coll") - pl.col("cooling_coll")).alias("collisional heating - cooling"),
        #     ["_yscale", "linear"],
        # ],
        # [['initmasses', ['Ni_56', 'He', 'C', 'Mg']]],
        # ['heating_gamma/gamma_dep'],
        # ["nne", ["_ymin", 1e5], ["_ymax", 1e10]],
        ["rho", ["_yscale", "log"], ["_ymin", 1e-16]],
        ["TR", ["_yscale", "linear"]],  # , ["_ymin", 1000], ["_ymax", 15000]
        # ["Te"],
        # ["Te", "TR"],
        [["averageionisation", ["Sr"]]],
        # [["averageexcitation", ["Fe II", "Fe III"]]],
        # [["populations", ["Sr90", "Sr91", "Sr92", "Sr94"]]],
        [["populations", ["Sr I", "Sr II", "Sr III", "Sr IV"]]],
        # [['populations', ['He I', 'He II', 'He III']]],
        # [['populations', ['C I', 'C II', 'C III', 'C IV', 'C V']]],
        # [['populations', ['O I', 'O II', 'O III', 'O IV']]],
        # [['populations', ['Ne I', 'Ne II', 'Ne III', 'Ne IV', 'Ne V']]],
        # [['populations', ['Si I', 'Si II', 'Si III', 'Si IV', 'Si V']]],
        # [['populations', ['Cr I', 'Cr II', 'Cr III', 'Cr IV', 'Cr V']]],
        # [['populations', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Fe VI', 'Fe VII', 'Fe VIII']]],
        # [['populations', ['Co I', 'Co II', 'Co III', 'Co IV', 'Co V', 'Co VI', 'Co VII']]],
        # [['populations', ['Ni I', 'Ni II', 'Ni III', 'Ni IV', 'Ni V', 'Ni VI', 'Ni VII']]],
        # [['populations', ['Fe II', 'Fe III', 'Co II', 'Co III', 'Ni II', 'Ni III']]],
        # [['populations', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Ni II']]],
        # [['gamma_NT', ['Fe I', 'Fe II', 'Fe III', 'Fe IV', 'Fe V', 'Ni II']]],
    ]

    estimatorcolumns = estimators.collect_schema().names()

    # detect a list of populations and convert to the correct form
    # e.g. ["Sr I", "Sr II"] is re-written to [['populations', ['Sr I', 'Sr II']]]
    for i in range(len(plotlist)):
        if isinstance(plotlist[i], str):
            plotlist[i] = [plotlist[i]]
        assert isinstance(plotlist[i], list)
        plot_directives = [
            plotvar.split("=", maxsplit=1)
            for plotvar in plotlist[i]
            if isinstance(plotvar, str) and plotvar.startswith("_") and "=" in plotvar
        ]
        plotlist[i] = [
            VARIABLE_ALIASES.get(plotvar, plotvar) if isinstance(plotvar, str) else plotvar
            for plotvar in plotlist[i]
            if not isinstance(plotvar, str) or not plotvar.startswith("_") or "=" not in plotvar
        ]
        if isinstance(plotlist[i][0], str) and plotlist[i][0] not in estimatorcolumns:
            # this is going to cause an error, so attempt to interpret it as populations
            rewrite_is_valid = False
            for plotvar in plotlist[i]:
                if isinstance(plotvar, list):
                    continue
                if not isinstance(plotvar, str):
                    break
                if plotvar.startswith("_"):
                    continue
                atomic_number, ionstage = get_iontuple(plotvar)
                if get_column_name("populations", atomic_number, ionstage)[0] not in estimatorcolumns:
                    break
            else:
                rewrite_is_valid = True
            if rewrite_is_valid:
                new_plotvars = [["populations", plotlist[i]]]
                print(f"Rewriting plotlist {plotlist[i]} to {new_plotvars}")
                plotlist[i] = new_plotvars
        if plot_directives:
            plotlist[i].extend(plot_directives)

    outdir = Path(args.outputfile) if Path(args.outputfile).is_dir() else Path()
    assert args.x is not None
    if args.x in {"time", "timestep"}:
        # plot time evolution
        make_figure(
            modelpath=modelpath,
            timestepslist=timesteps_included,
            estimators=estimators,
            xvariable=args.x,
            plotlist=plotlist,
            args=args,
        )
    else:
        # plot a range of cells in a time snapshot showing internal structure

        if args.x == "velocity" and modelmeta["vmax_cmps"] > 0.3 * 29979245800:
            args.x = "beta"

        if args.readonlymgi:
            if not isinstance(args.modelgridindex, list):
                args.modelgridindex = [args.modelgridindex] if args.modelgridindex is not None else []
            estimators = estimators.filter(pl.col("modelgridindex").is_in(args.modelgridindex))

        frames_timesteps_included = (
            [[ts] for ts in range(timestepmin, timestepmax + 1)] if args.multiplot else [timesteps_included]
        )

        if args.makegif:
            args.multiplot = True
            args.format = "png"

        outputfiles = []
        for timesteps_included in frames_timesteps_included:
            outfilename = make_figure(
                modelpath=modelpath,
                timestepslist=timesteps_included,
                estimators=estimators,
                xvariable=args.x,
                plotlist=plotlist,
                args=args,
            )

            outputfiles.append(outfilename)

        if len(outputfiles) > 1:
            if args.makegif:
                assert args.multiplot
                assert args.format == "png"
                import imageio.v2 as iio

                gifname = outdir / f"plotestim_evolution_ts{timestepmin:03d}_ts{timestepmax:03d}.gif"
                with iio.get_writer(gifname, mode="I", duration=1000) as writer:
                    for filename in outputfiles:
                        image = iio.imread(filename)
                        writer.append_data(image)  # type: ignore[attr-defined] # pyright: ignore[reportAttributeAccessIssue]
                print(f"Created gif: {gifname}")
            elif args.format == "pdf":
                at.merge_pdf_files(outputfiles)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
import sys

import numpy as np
import pandas as pd

import artistools as at


def get_theta_phi(anglebin: int) -> tuple[float | None, float | None]:
    """Get the central theta and phi angles for given anglebin."""
    assert isinstance(anglebin, int), "Anglebin has to be int"
    cos_theta = [-0.9, -0.7, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5, 0.7, 0.9]
    theta = np.arccos(cos_theta)
    phi = np.array([0.1, 0.3, 0.5, 0.7, 0.9, 1.9, 1.7, 1.5, 1.3, 1.1]) * np.pi

    anglenumber = 0
    for t in theta:
        for p in phi:
            if anglenumber == anglebin:
                return t, p

            anglenumber += 1

    return None, None


def gen_viewing_angle_df(length: int) -> pd.DataFrame:
    # Build viewing angle vector DataFrame
    viewing_angles: dict[str, list[float | str]] = {"Angle-bin": [], "x_coord": [], "y_coord": [], "z_coord": []}

    for i in range(100):
        theta, phi = get_theta_phi(i)
        assert theta is not None
        assert phi is not None
        x_c = length * np.sin(theta) * np.cos(phi)
        y_c = length * np.sin(theta) * np.sin(phi)
        z_c = length * np.cos(theta)

        # 0 point
        viewing_angles["Angle-bin"].append(f"{i:02d}")
        viewing_angles["x_coord"].append(0.0)
        viewing_angles["y_coord"].append(0.0)
        viewing_angles["z_coord"].append(0.0)

        # end point
        viewing_angles["Angle-bin"].append(f"{i:02d}")
        viewing_angles["x_coord"].append(x_c)
        viewing_angles["y_coord"].append(y_c)
        viewing_angles["z_coord"].append(z_c)

    return pd.DataFrame(viewing_angles)


def viewing_angles_visualisation(
    modelfile: str,
    outfile: str | None = None,
    isomin: float | None = None,
    isomax: float | None = None,
    opacity: float = 2.5,
    surface_count: int = 20,
    linewidth: float = 2.5,
    linelength: float = 1.0,
    show_plot: bool = False,
) -> tuple[float, float]:
    """Tool to generate a 3D visualization of an ARTIS model. Viewing angle bins will get overplotted with an animation.

    Parameters
    ----------
    modelfile : str
        File where ARTIS  model is stored.
    outfile : str
        Name of the output file. If name contains 'html',
        figure will be stored as html file including
        the animation
    isomin : float
        Minimum density value for the color coding
    isomax : float
        Maximum density value for the color coding
    opacity : float
        Opacity value
    surface_count : int
        Number of isosurfaces plotted
    linewidth : float
        Width of the viewing angle lines
    linelength : float
        Length of the viewing angle lines in units
        of the boxsize
    show_plot : bool
        If True, plot will be shown after saving

    Returns
    -------
    isomin, isomax : float, float

    """
    try:
        import plotly.express as px
        import plotly.graph_objects as go
    except ModuleNotFoundError:
        print("Cannot run visualization without plotly...")
        sys.exit()

    # Load model contents
    dfmodel = (
        at.get_modeldata(modelfile, derived_cols=["pos_mid"])[0].collect().to_pandas(use_pyarrow_extension_array=True)
    )
    x, y, z = (dfmodel[f"pos_{ax}_mid"].to_numpy(dtype=float) for ax in ("x", "y", "z"))
    rho = dfmodel["rho"].to_numpy(dtype=float)

    if isomin is None:
        isomin = min(rho.flatten())
    if isomax is None:
        isomax = max(rho.flatten())
    assert isomin is not None
    assert isomax is not None
    assert isomin < isomax, "isomin must be smaller than isomax"

    # Generate viewing angle vectory
    length = max(x.flatten()) * linelength
    va = gen_viewing_angle_df(length)

    # Create plot
    fig = px.line_3d(
        va,
        x="x_coord",
        y="y_coord",
        z="z_coord",
        color="Angle-bin",
        animation_frame="Angle-bin",
        hover_name="Angle-bin",
    )
    fig.update_traces(line={"width": linewidth})
    fig.update_layout(legend={"orientation": "h", "yanchor": "bottom", "y": 1.02, "xanchor": "right", "x": 1})

    fig = fig.add_trace(
        go.Volume(
            x=x.flatten(),
            y=y.flatten(),
            z=z.flatten(),
            value=rho.flatten(),
            isomin=isomin,
            isomax=isomax,
            opacity=opacity,  # needs to be small to see through all surfaces
            surface_count=surface_count,  # needs to be a large number for good volume rendering
            colorbar={"title": "Density (g/cm³)"},
        )
    )
    fig.update_layout(
        scene_xaxis_showticklabels=False, scene_yaxis_showticklabels=False, scene_zaxis_showticklabels=False
    )

    if outfile:
        if outfile.endswith("html"):
            fig.write_html(outfile, auto_play=False)
        else:
            fig.write_image(outfile)
        print(f"Figure saved as {outfile}")

    if show_plot:
        fig.show()

    return (isomin, isomax)


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("modelfile", help="Path to the ARTIS model.")
    parser.add_argument(
        "-o",
        "--outfile",
        help="Name of the output file. If it contains 'html', figure will be stored as html including the animation.",
    )
    parser.add_argument("--isomin", type=float, help="Minimum density for color coding.")
    parser.add_argument("--isomax", type=float, help="Maximum density for color coding.")
    parser.add_argument("--opacity", type=float, help="Opacity value. Default: 0.25", default=0.25)
    parser.add_argument(
        "-s", "--surface_count", type=int, help="Number of isosurfaces plotted. Default: 20", default=20
    )
    parser.add_argument("--linewidth", type=float, help="Width of the viewing angle lines. Default: 2.5", default=2.5)
    parser.add_argument(
        "--linelength",
        type=float,
        help="Length of the viewing angle lines in units of the boxsize. Default: 1.0",
        default=1.0,
    )
    parser.add_argument("--show_plot", action="store_true", help="If flag is given, plot will be shown after saving.")


def main() -> None:
    """Tool to generate a 3D visualization of an ARTIS model."""
    parser = argparse.ArgumentParser()

    addargs(parser)

    args = parser.parse_args()

    viewing_angles_visualisation(
        modelfile=args.modelfile,
        outfile=args.outfile,
        isomin=args.isomin,
        isomax=args.isomax,
        opacity=args.opacity,
        surface_count=args.surface_count,
        linewidth=args.linewidth,
        linelength=args.linelength,
        show_plot=args.show_plot,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# adapted from Fortran maptogrid.f90 and kernelmodule.f90
# original Fortran code by Andreas Bauswein

import argparse
import math
import typing as t
from collections.abc import Sequence
from pathlib import Path

import argcomplete
import numpy as np
import numpy.typing as npt
import polars as pl

import artistools as at

itable = 40000  # wie fein Kernelfkt interpoliert wird
itab = itable + 5

#
# --maximum interaction length and step size
#
v2max = 4.0
dvtable = v2max / itable
i1 = int(1.0 // dvtable)


def get_wij() -> npt.NDArray[np.floating]:
    igphi = 0
    #
    # --normalisation constant
    #
    cnormk = 1.0 / math.pi
    # --build tables
    #
    #  a) v less than 1
    #
    wij = np.zeros(itab + 1)
    if igphi == 1:
        for i in range(1, i1 + 1):
            v2 = i * dvtable
            v = math.sqrt(v2)
            v3 = v * v2
            # v4 = v * v3
            vsum = 1.0 - 1.5 * v2 + 0.75 * v3
            wij[i] = cnormk * vsum
    else:
        for i in range(1, i1 + 1):
            v2 = i * dvtable
            v = math.sqrt(v2)
            v3 = v * v2
            vsum = 1.0 - 1.5 * v2 + 0.75 * v3
            wij[i] = cnormk * vsum

    #
    #  b) v greater than 1
    #
    if igphi == 1:
        for i in range(i1 + 1, itable + 1):
            v2 = i * dvtable
            v = math.sqrt(v2)
            dif2 = 2.0 - v
            vsum = 0.25 * dif2 * dif2 * dif2
            wij[i] = cnormk * vsum
    else:
        for i in range(i1 + 1, itable + 1):
            v2 = i * dvtable
            v = math.sqrt(v2)
            dif2 = 2.0 - v
            vsum = 0.25 * dif2 * dif2 * dif2
            wij[i] = cnormk * vsum

    return wij


def kernelvals2(
    rij2: float, hmean: float, wij: npt.NDArray[np.floating]
) -> float:  # ist schnell berechnet aber keine Gradienten
    hmean21 = 1.0 / hmean**2
    hmean31 = hmean21 / hmean
    v2 = rij2 * hmean21
    index = math.floor(v2 / dvtable)
    dxx = v2 - index * dvtable
    index1 = index + 1
    dwdx = (wij[index1] - wij[index]) / dvtable
    val = (wij[index] + dwdx * dxx) * hmean31
    assert isinstance(val, float)
    return val


def maptogrid(
    ejectasnapshotpath: Path,
    outputfolderpath: Path | str,
    ncoordgrid: int = 50,
    downsamplefactor: int = 1,
    dtextra_seconds: float = 0.5,
    setgrid_fractionrmax: float = 0.5,
    modifysmoothinglength: str = "option4",
) -> None:
    if not ejectasnapshotpath.is_file():
        print(f"{ejectasnapshotpath} not found")
        raise FileNotFoundError

    outputfolderpath = Path(outputfolderpath)
    if not outputfolderpath.exists():
        outputfolderpath.mkdir(parents=True)

    # save the printed output to a log file
    logprint = at.inputmodel.inputmodel_misc.savetologfile(
        outputfolderpath=outputfolderpath, logfilename="maptogridlog.txt"
    )

    wij = get_wij()

    assert ncoordgrid % 2 == 0

    snapshot_columns_used = ["id", "h", "x", "y", "z", "vx", "vy", "vz", "pmass", "rho", "p", "rho_rst", "ye"]

    dfsnapshot = at.inputmodel.modelfromhydro.read_ejectasnapshot(
        ejectasnapshotpath, usecols=snapshot_columns_used, downsamplefactor=downsamplefactor
    )

    logprint(dfsnapshot)
    logprint(f"ncoordgrid: {ncoordgrid}")

    assert len(dfsnapshot.columns) == len(snapshot_columns_used)

    npart = len(dfsnapshot)

    # Propagate particles to dtextra using velocities
    logprint(f"Propagating particles for dtextra_seconds={dtextra_seconds}")
    dtextra = dtextra_seconds / 4.926e-6  # convert to geom units.
    dfsnapshot = (
        dfsnapshot.with_columns(
            dis_orig=(pl.col("x") ** 2 + pl.col("y") ** 2 + pl.col("z") ** 2).sqrt(),
            x_orig=pl.col("x"),
            y_orig=pl.col("y"),
            z_orig=pl.col("z"),
            x=pl.col("x") + pl.col("vx") * dtextra,
            y=pl.col("y") + pl.col("vy") * dtextra,
            z=pl.col("z") + pl.col("vz") * dtextra,
        )
        .with_columns(dis=(pl.col("x") ** 2 + pl.col("y") ** 2 + pl.col("z") ** 2).sqrt())
        .with_columns(
            h=pl.col("h") / pl.col("dis_orig") * pl.col("dis"),
            vrad=(pl.col("vx") * pl.col("x") + pl.col("vy") * pl.col("y") + pl.col("vz") * pl.col("z")) / pl.col("dis"),
            vtot=(pl.col("vx") ** 2 + pl.col("vy") ** 2 + pl.col("vz") ** 2).sqrt(),
        )
        .with_columns(
            vperp=pl.when(pl.col("vtot") > pl.col("vrad"))
            .then((pl.col("vtot") ** 2 - pl.col("vrad") ** 2).sqrt())
            .otherwise(0.0)
        )
    )

    particleid = dfsnapshot["id"].to_numpy()
    x = dfsnapshot["x"].to_numpy()
    y = dfsnapshot["y"].to_numpy()
    z = dfsnapshot["z"].to_numpy()
    h = dfsnapshot["h"].to_numpy().copy()
    pmass = dfsnapshot["pmass"].to_numpy()
    rho_rst = dfsnapshot["rho_rst"].to_numpy()
    rho = dfsnapshot["rho"].to_numpy()
    Ye = dfsnapshot["ye"].to_numpy()

    totmass = dfsnapshot["pmass"].sum()
    rmean = dfsnapshot["dis"].mean()
    assert isinstance(rmean, float)
    hmean = dfsnapshot["h"].mean()
    assert isinstance(hmean, float)
    hmin = dfsnapshot["h"].min()
    assert isinstance(hmin, float)
    rmax = dfsnapshot["dis"].max()
    assert isinstance(rmax, float)
    with Path(outputfolderpath, "ejectapartanalysis.dat").open(mode="w", encoding="utf-8") as fpartanalysis:
        fpartanalysis.writelines(
            f"{part['dis']} {part['h']} {part['h'] / part['dis']} {part['vrad']} {part['vperp']} {part['vtot']}\n"
            for part in dfsnapshot.select(["dis", "h", "vrad", "vperp", "vtot"]).iter_rows(named=True)
        )

    logprint(f"saved {outputfolderpath / 'ejectapartanalysis.dat'}")

    logprint(f"total mass of sph particle {totmass} max dist {rmax} mean dist {rmean}")
    logprint(f"smoothing length min {hmin} mean {hmean}")
    logprint("ratio between vrad and vperp mean", dfsnapshot.select(pl.col("vperp") - pl.col("vrad")).mean().item(0, 0))

    # check maybe cm and correct by shifting

    # ...

    # set up grid
    logprint(
        f"setgrid_fractionrmax={setgrid_fractionrmax}: gridmax is set to {setgrid_fractionrmax}*rmax of the SPH particles"
    )
    x0 = -setgrid_fractionrmax * rmax  # Set x0 (gridmax) to a fraction of the maximum radius of the SPH particles
    # Default is 50% (but is hand waving - choose) #
    # x0 = - rmean

    dx = 2.0 * abs(x0) / (ncoordgrid)  # -1 to be symmetric, right?

    y0 = x0
    z0 = x0
    dy = dx
    dz = dx

    grho = np.zeros((ncoordgrid, ncoordgrid, ncoordgrid))
    gye = np.zeros((ncoordgrid, ncoordgrid, ncoordgrid))
    gparticlecounter = np.zeros((ncoordgrid, ncoordgrid, ncoordgrid), dtype=int)
    particle_rho_contribs = {}

    logprint(f"grid properties {x0=}, {dx=}, {x0 + dx * (ncoordgrid - 1)=}")

    arrgx = x0 + dx * np.arange(ncoordgrid)
    arrgy = arrgx
    arrgz = arrgx

    particlesused = set()
    particlesinsidegrid = set()

    logprint(f"modifysmoothinglength: {modifysmoothinglength}")

    for n in range(npart):
        maxdist = 2.0 * h[n]
        maxdist2 = maxdist**2

        ilow = max(math.floor((x[n] - maxdist - x0) / dx), 0)
        ihigh = min(math.ceil((x[n] + maxdist - x0) / dx), ncoordgrid - 1)
        jlow = max(math.floor((y[n] - maxdist - y0) / dy), 0)
        jhigh = min(math.ceil((y[n] + maxdist - y0) / dy), ncoordgrid - 1)
        klow = max(math.floor((z[n] - maxdist - z0) / dz), 0)
        khigh = min(math.ceil((z[n] + maxdist - z0) / dz), ncoordgrid - 1)

        if min(ihigh, jhigh, khigh) >= 1 and max(ilow, jlow, klow) <= ncoordgrid:
            particlesinsidegrid.add(n)
        # check some min max

        # ... kernel reweighting ?

        searchcoords = [
            (i, j, k, (arrgx[i] - x[n]) ** 2 + (arrgy[j] - y[n]) ** 2 + (arrgz[k] - z[n]) ** 2)
            for i in range(ilow, ihigh + 1)
            for j in range(jlow, jhigh + 1)
            for k in range(klow, khigh + 1)
        ]

        for i, j, k, dis2 in searchcoords:
            if modifysmoothinglength != "False":
                # -- change h by hand --------- we could do these particle thinsg also further up

                # option 1 minimum that no particle is lost

                # option 2 increase smoothing everywhere, i.e. less holes but also less structure

                # option 3 increase smoothing beyond some distance

                # options can be combined, i.e. option 1 alone fills the hole in the center
                # (which we could also replace by later ejecta)
                if modifysmoothinglength == "option1":
                    h[n] = max(h[n], 1.5 * dx)  # option 1

                dis = math.sqrt(x[n] * x[n] + y[n] * y[n] + z[n] * z[n])

                if modifysmoothinglength == "option2":
                    h[n] = max(h[n], 0.25 * dis)  # option 2

                if modifysmoothinglength == "option3" and dis > 1.5 * rmean:
                    h[n] = max(h[n], 0.4 * dis)  # option 3

                # option 4 (default) -- for particles with radius > mean particle radius choose the larger h
                # from the particle h and 150% of the mean h for all particles
                if modifysmoothinglength == "option4" and dis > rmean:
                    h[n] = max(h[n], hmean * 1.5)

                maxdist2 = (2.0 * h[n]) ** 2
                # -------------------------------

                # or via neighbors  - not yet implemented

            if dis2 <= maxdist2:
                wtij = kernelvals2(dis2, float(h[n]), wij)

                # USED PREVIOUSLY: less accurate?
                # grho_contrib = pmass[n] * wtij

                # this particle's contribution to mass density (rho) in the cell
                grho_contrib = pmass[n] * rho[n] / rho_rst[n] * wtij

                grho[i, j, k] += grho_contrib

                particle_rho_contribs[n, i, j, k] = grho_contrib

                # mass-weighted electron fraction (needs to be normalised by cell density afterwards)
                gye[i, j, k] += grho_contrib * Ye[n]

                # count number of particles contributing to each grid cell
                gparticlecounter[i, j, k] += 1
                particlesused.add(n)

    logprint(
        f"particles with any cell contribution: {len(particlesused)} of {len(particlesinsidegrid)} inside grid out of"
        f" {npart} total"
    )
    unusedparticles = [n for n in range(npart) if n not in particlesused]
    for n in unusedparticles:
        loc_i = math.floor((x[n] - x0) / dx)
        loc_j = math.floor((y[n] - y0) / dy)
        loc_k = math.floor((z[n] - z0) / dz)
        # ignore particles outside grid boundary
        if min(loc_i, loc_j, loc_k) < 0 or max(loc_i, loc_j, loc_k) > ncoordgrid - 1:
            continue
        logprint(f"particle {n} is totally unused but located in cell {loc_i} {loc_j} {loc_k}")

    with np.errstate(divide="ignore", invalid="ignore"):
        gye = np.divide(gye, grho)

        with Path(outputfolderpath, "gridcontributions.txt").open("w", encoding="utf-8") as fcontribs:
            fcontribs.write("particleid cellindex frac_of_cellmass\n")
            for (n, i, j, k), rho_contrib in particle_rho_contribs.items():
                gridindex = (k * ncoordgrid + j) * ncoordgrid + i + 1
                fcontribs.write(f"{particleid[n]} {gridindex} {rho_contrib / grho[i, j, k]}\n")
        logprint(f"saved {outputfolderpath / 'gridcontributions.txt'}")

    # check some stuff on the grid

    nzero = 0
    nzerocentral = 0
    gmass = np.sum(grho) * dx * dy * dz
    # nzero = np.count_nonzero(grho[1:][1:][1:] < 1.e-20)

    for i in range(ncoordgrid):
        gx = x0 + dx * i
        for j in range(ncoordgrid):
            gy = y0 + dy * j
            for k in range(1, ncoordgrid):
                # how many cells with rho=0?

                if grho[i, j, k] < 1.0e-20:
                    nzero += 1

                gz = z0 + dz * k

                dis = math.sqrt(gx * gx + gy * gy + gz * gz)

                if grho[i, j, k] < 1.0e-20 and dis < rmean:
                    nzerocentral += 1

    logprint(f"fraction of total mass on grid {gmass / totmass}")

    logprint(
        f"{'WARNING!' if gmass / totmass < 0.9 else ''} mass on grid from rho*V: {gmass} mass of particles: {totmass} "
    )

    logprint(
        f"number of cells with rho=0 {nzero}, total num of cells {ncoordgrid**3}, fraction of cells with rho=0:"
        f" {(nzero) / (ncoordgrid**3)}"
    )

    logprint(
        f"number of central cells (dis<rmean) with rho=0 {nzerocentral}, ratio"
        f" {(nzerocentral) / (4.0 * math.pi / 3.0 * rmean**3 / (dx * dy * dz))}"
    )

    logprint("probably we want to choose grid size, i.e. x0, as compromise between mapped mass and rho=0 cells")

    # output grid - adapt as you need output

    with Path(outputfolderpath, "grid.dat").open("w", encoding="utf-8") as fgrid:
        fgrid.write(f"{ncoordgrid**3} # ncoordgrid\n")
        fgrid.write(f"{dtextra} # extra time after explosion simulation ended (in geom units)\n")
        fgrid.write(f"{x0} # xmax\n")
        fgrid.write(" gridindex    pos_x_min    pos_y_min    pos_z_min    rho    cellYe    tracercount\n")
        gridindex = 1
        for k in range(ncoordgrid):
            gz = z0 + dz * k
            for j in range(ncoordgrid):
                gy = y0 + dy * j
                for i in range(ncoordgrid):
                    fgrid.write(
                        f"{gridindex:8d} {x0 + dx * i} {gy} {gz} {grho[i, j, k]} {gye[i, j, k]} {gparticlecounter[i, j, k]}\n"
                    )
                    # gridindex2 = ((k - 1) * ncoordgrid + (j - 1)) * ncoordgrid + (i - 1) + 1

                    gridindex += 1

    logprint(f"saved {outputfolderpath / 'grid.dat'}")


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-inputpath", "-i", default=".", help="Path to ejectasnapshot")
    parser.add_argument(
        "-ncoordgrid", type=int, default=50, help="Number of grid positions per axis (numcells = ncoordgrid^3)"
    )
    parser.add_argument(
        "-dtextra_seconds",
        type=float,
        default=0.5,
        help="Time in seconds to propagate SPH particles ballistically after end of SPH simulation."
        " 0 for no extrapolation",
    )
    parser.add_argument(
        "-setgrid_fractionrmax",
        type=float,
        default=0.5,
        help="Setup grid to have max equal to fraction of particle rmax.",
    )
    parser.add_argument(
        "-downsamplefactor",
        type=int,
        default=1,
        help="Randomly sample particles, reducing the number by this factor (e.g. 2 will ignore half of the particles)",
    )
    parser.add_argument(
        "-modifysmoothinglength",
        default="option4",
        choices=[
            "option1",
            "option2",
            "option3",
            "option4",
            "False",
        ],  # We should choose if the default should be false and how we want to name these
        help="Option to modify smoothing length h. Choose from options."
        "Default modifies h. Set to False for no modifications to h.",
    )

    parser.add_argument("-outputpath", "-o", default=".", help="Path for output files")


def main(args: argparse.Namespace | None = None, argsraw: Sequence[str] | None = None, **kwargs: t.Any) -> None:
    """Map tracer particle trajectories to a Cartesian grid."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    ejectasnapshotpath = Path(args.inputpath, "ejectasnapshot.dat")

    maptogrid(
        ejectasnapshotpath=ejectasnapshotpath,
        ncoordgrid=args.ncoordgrid,
        outputfolderpath=args.outputpath,
        downsamplefactor=args.downsamplefactor,
        dtextra_seconds=args.dtextra_seconds,
        setgrid_fractionrmax=args.setgrid_fractionrmax,
        modifysmoothinglength=args.modifysmoothinglength,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Functions for reading and processing estimator files.

Examples are temperatures, populations, and heating/cooling rates.
"""

import contextlib
import datetime
import tempfile
import time
import typing as t
from collections.abc import Collection
from collections.abc import Sequence
from pathlib import Path

import polars as pl
from polars import selectors as cs

import artistools as at


def get_variableunits(key: str) -> str | None:
    variableunits = {
        "time": "days",
        "gamma_NT": "s^-1",
        "gamma_R_bfest": "s^-1",
        "TR": "K",
        "Te": "K",
        "TJ": "K",
        "nne": "e$^-$/cm$^3$",
        "nniso": "cm$^{-3}$",
        "nnion": "cm$^{-3}$",
        "nnelement": "cm$^{-3}$",
        "deposition": "erg/s/cm$^3$",
        "total_dep": "erg/s/cm$^3$",
        "heating": "erg/s/cm$^3$",
        "heating_dep/total_dep": "Ratio",
        "cooling": "erg/s/cm$^3$",
        "rho": "g/cm$^3$",
        "velocity": "km/s",
        "beta": "v/c",
        "vel_r_max_kmps": "km/s",
        **{f"vel_{ax}_mid": "cm/s" for ax in ["x", "y", "z", "r", "rcyl"]},
        **{f"vel_{ax}_mid_on_c": "c" for ax in ["x", "y", "z", "r", "rcyl"]},
    }

    return variableunits.get(key) or variableunits.get(key.split("_", maxsplit=1)[0])


def get_variablelongunits(key: str) -> str | None:
    return {"heating_dep/total_dep": "", "TR": "Temperature [K]", "Te": "Temperature [K]", "TJ": "Temperature [K]"}.get(
        key
    )


def get_varname_formatted(varname: str) -> str:
    return {
        "nne": r"n$_{\rm e}$",
        "lognne": r"Log n$_{\rm e}$",
        "rho": r"$\rho$",
        "Te": r"T$_{\rm e}$",
        "TR": r"T$_{\rm R}$",
        "TJ": r"T$_{\rm J}$",
        "gamma_NT": r"$\Gamma_{\rm non-thermal}$ [s$^{-1}$]",
        "gamma_R_bfest": r"$\Gamma_{\rm phot}$ [s$^{-1}$]",
        "heating_dep/total_dep": "Heating fraction",
        **{f"vel_{ax}_mid": f"$v_{{{ax}}}$" for ax in ["x", "y", "z", "r", "rcyl"]},
        **{f"vel_{ax}_mid_on_c": f"$v_{{{ax}}}$" for ax in ["x", "y", "z", "r", "rcyl"]},
    }.get(varname, varname)


def get_units_string(variable: str) -> str:
    return f" [{units}]" if (units := get_variableunits(variable)) else ""


def get_rankbatch_parquetfile(
    folderpath: Path | str,
    batch_mpiranks: Sequence[int],
    batchindex: int,
    modelpath: Path | str | None = None,
    verbose: bool = True,
) -> Path:
    printornot = print if verbose else lambda _: None
    modelpath = Path(folderpath).parent if modelpath is None else Path(modelpath)
    folderpath = Path(folderpath)
    parquetfilename = f"estimbatch{batchindex:02d}_{batch_mpiranks[0]:04d}_{batch_mpiranks[-1]:04d}.out.parquet.tmp"
    parquetfilepath = folderpath / parquetfilename

    textsource_mtime: float | None = None
    with contextlib.suppress(StopIteration):
        textsource_mtime = next(folderpath.glob("estimators_????.out*")).stat().st_mtime

    assert len(batch_mpiranks) == max(batch_mpiranks) - min(batch_mpiranks) + 1, (
        "batch_mpiranks must be a contiguous range of ranks"
    )
    assert len(set(batch_mpiranks)) == len(batch_mpiranks), "batch_mpiranks must not contain duplicates"

    if not parquetfilepath.exists():
        generate_parquet = True
    elif textsource_mtime and textsource_mtime > parquetfilepath.stat().st_mtime:
        print(
            f"  {parquetfilepath.relative_to(modelpath.parent)} is older than the estimator text files. File will be deleted and regenerated..."
        )
        parquetfilepath.unlink()
        generate_parquet = True
    else:
        generate_parquet = False

    if generate_parquet:
        print(f"  generating {parquetfilepath.relative_to(modelpath.parent)}...")

        time_start = time.perf_counter()

        print(
            f"    reading {len(batch_mpiranks)} estimator files in {folderpath.relative_to(Path(folderpath).parent)}...",
            end="",
            flush=True,
        )

        pldf_batch = pl.DataFrame(at.rustext.estimparse(str(folderpath), min(batch_mpiranks), max(batch_mpiranks)))

        pldf_batch = pldf_batch.with_columns(
            cs.by_name("titeration", "timestep", "modelgridindex", require_all=False).cast(pl.Int32)
        )

        pldf_batch = pldf_batch.select(
            sorted(
                pldf_batch.columns,
                key=lambda col: f"-{col!r}" if col in {"timestep", "modelgridindex", "titer"} else str(col),
            )
        )
        print(f"took {time.perf_counter() - time_start:.1f} s. Writing parquet file...", end="", flush=True)
        time_start = time.perf_counter()

        assert pldf_batch is not None
        partialparquetfilepath = Path(
            tempfile.mkstemp(dir=folderpath, prefix=f"{parquetfilename}.partial", suffix=".partial")[1]
        )
        pldf_batch.write_parquet(
            partialparquetfilepath,
            compression="zstd",
            compression_level=10,
            statistics=True,
            metadata={
                "creationtimeutc": str(datetime.datetime.now(datetime.UTC)),
                "textsource_mtime": str(textsource_mtime),
                "batch_rank_min": str(min(batch_mpiranks)),
                "batch_rank_max": str(max(batch_mpiranks)),
                "batchindex": str(batchindex),
            },
        )
        if parquetfilepath.exists():
            partialparquetfilepath.unlink()
        else:
            partialparquetfilepath.rename(parquetfilepath)

        print(f"took {time.perf_counter() - time_start:.1f} s.")

    filesize = parquetfilepath.stat().st_size / 1024 / 1024
    try:
        printornot(f"  scanning {parquetfilepath.relative_to(modelpath.parent)} ({filesize:.2f} MiB)")
    except ValueError:
        printornot(f"  scanning {parquetfilepath} ({filesize:.2f} MiB)")

    return parquetfilepath


def join_cell_modeldata(
    estimators: pl.LazyFrame, modelpath: Path | str, verbose: bool = False
) -> tuple[pl.LazyFrame, dict[str, t.Any]]:
    """Join the estimator data with data from model.txt and derived quantities, e.g. density, volume, etc."""
    assert estimators is not None
    estimators = estimators.join(
        at.get_timesteps(modelpath).select("timestep", "tmid_days", "twidth_days"),
        on="timestep",
        how="left",
        coalesce=True,
    )
    dfmodel, modelmeta = at.inputmodel.get_modeldata(
        modelpath, derived_cols=["ALL"], get_elemabundances=True, printwarningsonly=not verbose
    )

    dfmodel = dfmodel.rename({
        colname: f"init_{colname}"
        for colname in dfmodel.collect_schema().names()
        if not colname.startswith("vel_") and colname not in {"inputcellid", "modelgridindex", "mass_g"}
    })
    return estimators.join(dfmodel, on="modelgridindex", suffix="_initmodel").with_columns(
        rho=pl.col("init_rho") * (modelmeta["t_model_init_days"] / pl.col("tmid_days")) ** 3,
        volume=pl.col("init_volume") * (pl.col("tmid_days") / modelmeta["t_model_init_days"]) ** 3,
    ), modelmeta


def scan_estimators(
    modelpath: Path | str = Path(),
    modelgridindex: int | Sequence[int] | None = None,
    timestep: int | Sequence[int] | None = None,
    join_modeldata: bool = False,
    verbose: bool = True,
) -> pl.LazyFrame:
    """Read estimator files into a dictionary of (timestep, modelgridindex): estimators.

    Selecting particular timesteps or modelgrid cells will using speed this up by reducing the number of files that must be read.
    """
    modelpath = Path(modelpath)
    match_modelgridindex: Sequence[int] | None
    if modelgridindex is None:
        match_modelgridindex = None
    elif isinstance(modelgridindex, int):
        match_modelgridindex = (modelgridindex,)
    else:
        match_modelgridindex = tuple(modelgridindex)

    match_timestep: Sequence[int] | None
    if timestep is None:
        match_timestep = None
    elif isinstance(timestep, int):
        match_timestep = (timestep,)
    else:
        match_timestep = tuple(timestep)

    if not Path(modelpath).exists() and Path(modelpath).parts[0] == "codecomparison":
        estimators = at.codecomparison.read_reference_estimators(
            modelpath, timestep=timestep, modelgridindex=modelgridindex
        )
        return pl.DataFrame([
            {"timestep": ts, "modelgridindex": mgi, **estimvals} for (ts, mgi), estimvals in estimators.items()
        ]).lazy()

    # print(f" matching cells {match_modelgridindex} and timesteps {match_timestep}")
    mpiranklist = at.get_mpiranklist(modelpath, only_ranks_withgridcells=True)
    mpiranks_matched = (
        {at.get_mpirankofcell(modelpath=modelpath, modelgridindex=mgi) for mgi in match_modelgridindex}
        if match_modelgridindex
        else set(mpiranklist)
    )
    mpirank_groups = [
        (batchindex, mpiranks)
        for batchindex, mpiranks in enumerate(at.misc.batched(mpiranklist, 100))
        if mpiranks_matched.intersection(mpiranks)
    ]

    runfolders = at.get_runfolders(modelpath, timesteps=match_timestep)
    if runfolders:
        parquetfiles = (
            get_rankbatch_parquetfile(
                modelpath=modelpath,
                folderpath=runfolder,
                batch_mpiranks=mpiranks,
                batchindex=batchindex,
                verbose=verbose,
            )
            for runfolder in runfolders
            for batchindex, mpiranks in mpirank_groups
        )

        assert bool(parquetfiles)

        pldflazy = (
            pl.concat([pl.scan_parquet(pfile) for pfile in parquetfiles], how="diagonal_relaxed")
            .unique(["timestep", "modelgridindex"], maintain_order=True, keep="first")
            .lazy()
        )
    else:
        print(f"WARNING: No run folders found in {modelpath}. Enabling fallback to cross join of all cells/timesteps.")
        pldflazy = (
            at.get_timesteps(modelpath)
            .select("timestep", "tmid_days", "twidth_days")
            .join(at.inputmodel.get_modeldata(modelpath)[0].select("modelgridindex"), how="cross")
        )

    if match_modelgridindex is not None:
        pldflazy = pldflazy.filter(pl.col("modelgridindex").is_in(match_modelgridindex))

    if match_timestep is not None:
        pldflazy = pldflazy.filter(pl.col("timestep").is_in(match_timestep))

    colnames = pldflazy.collect_schema().names()
    # add some derived quantities
    if "heating_gamma/gamma_dep" in colnames:
        pldflazy = pldflazy.with_columns(gamma_dep=pl.col("heating_gamma") / pl.col("heating_gamma/gamma_dep"))

    if "deposition_gamma" in colnames:
        # sum up the gamma, elec, positron, alpha deposition contributions
        pldflazy = pldflazy.with_columns(total_dep=pl.sum_horizontal(cs.starts_with("deposition_")))
    elif "heating_heating_dep/total_dep" in colnames:
        # for older files with no deposition data, take heating part of deposition and heating fraction
        pldflazy = pldflazy.with_columns(total_dep=pl.col("heating_dep") / pl.col("heating_heating_dep/total_dep"))

    if any(col.startswith("nnelement_") for col in colnames):
        pldflazy = pldflazy.with_columns(nntot=pl.sum_horizontal(cs.starts_with("nnelement_"))).fill_null(0)

    if join_modeldata:
        pldflazy, _ = join_cell_modeldata(estimators=pldflazy, modelpath=modelpath, verbose=verbose)

    return pldflazy


def read_estimators(
    modelpath: Path | str = Path(),
    modelgridindex: int | Sequence[int] | None = None,
    timestep: int | Sequence[int] | None = None,
    keys: Collection[str] | None = None,
) -> dict[tuple[int, int], dict[str, t.Any]]:
    """Read ARTIS estimator data into a dictionary keyed by (timestep, modelgridindex).

    When collecting many cells and timesteps, this is very slow, and it's almost always better to use scan_estimators instead.
    """
    if isinstance(keys, str):
        keys = {keys}
    lzpldfestimators = scan_estimators(modelpath, modelgridindex, timestep)

    if isinstance(modelgridindex, int):
        lzpldfestimators = lzpldfestimators.filter(pl.col("modelgridindex") == modelgridindex)
    elif isinstance(modelgridindex, Sequence):
        lzpldfestimators = lzpldfestimators.filter(pl.col("modelgridindex").is_in(modelgridindex))
    if isinstance(timestep, int):
        lzpldfestimators = lzpldfestimators.filter(pl.col("timestep") == timestep)
    elif isinstance(timestep, Sequence):
        lzpldfestimators = lzpldfestimators.filter(pl.col("timestep").is_in(timestep))

    pldfestimators = lzpldfestimators.collect()

    estimators: dict[tuple[int, int], dict[str, t.Any]] = {}
    for estimtsmgi in pldfestimators.iter_rows(named=True):
        ts, mgi = estimtsmgi["timestep"], estimtsmgi["modelgridindex"]
        estimators[ts, mgi] = {
            k: v
            for k, v in estimtsmgi.items()
            if k not in {"timestep", "modelgridindex"} and (keys is None or k in keys) and v is not None
        }

    return estimators


def get_averageexcitation(
    modelpath: Path | str, modelgridindex: int, timestep: int, atomic_number: int, ion_stage: int, T_exc: float
) -> float | None:
    dfnltepops = at.nltepops.read_files(modelpath, modelgridindex=modelgridindex, timestep=timestep)
    if dfnltepops.empty:
        print(f"WARNING: NLTE pops not found for cell {modelgridindex} at timestep {timestep}")

    adata = at.atomic.get_levels(modelpath)
    ionlevels = adata.filter((pl.col("Z") == atomic_number) & (pl.col("ion_stage") == ion_stage))["levels"].item()

    energypopsum = 0
    ionpopsum = 0
    if dfnltepops.empty:
        return None

    dfnltepops_ion = dfnltepops.query(
        "modelgridindex==@modelgridindex and timestep==@timestep and Z==@atomic_number & ion_stage==@ion_stage"
    )

    k_b = 8.617333262145179e-05  # eV / K

    ionpopsum = dfnltepops_ion.n_NLTE.sum()
    assert isinstance(ionpopsum, float)
    energypopsum = sum(
        ionlevels["energy_ev"].item(level) * n_NLTE
        for level, n_NLTE in dfnltepops_ion[dfnltepops_ion.level >= 0][["level", "n_NLTE"]].itertuples(index=False)
    )
    assert isinstance(energypopsum, float)

    with contextlib.suppress(IndexError):  # no superlevel will cause IndexError
        superlevelrow = dfnltepops_ion[dfnltepops_ion.level < 0].iloc[0]
        levelnumber_sl = dfnltepops_ion.level.max() + 1

        energy_boltzfac_sum = (
            ionlevels[levelnumber_sl:]
            .select(pl.col("energy_ev") * pl.col("g") * (-pl.col("energy_ev") / k_b / T_exc).exp())
            .sum()
            .item()
        )

        boltzfac_sum = energy_boltzfac_sum = (
            ionlevels[levelnumber_sl:].select(pl.col("g") * (-pl.col("energy_ev") / k_b / T_exc).exp()).sum().item()
        )
        # adjust to the actual superlevel population from ARTIS
        energypopsum += energy_boltzfac_sum * superlevelrow.n_NLTE / boltzfac_sum

    return energypopsum / ionpopsum



#!/usr/bin/env python3
import typing as t
from unittest import mock

import matplotlib.axes as mplax
import numpy as np
import pytest
from pytest_codspeed.plugin import BenchmarkFixture

import artistools as at


@mock.patch.object(mplax.Axes, "plot", side_effect=mplax.Axes.plot, autospec=True)
@pytest.mark.benchmark
def test_vspectraplot(mockplot: t.Any, benchmark: BenchmarkFixture) -> None:
    at.spectra.plot(
        argsraw=[],
        specpath=[at.get_config()["path_testdata"] / "vspecpolmodel", "sn2011fe_PTF11kly_20120822_norm.txt"],
        outputfile=at.get_config()["path_testoutput"] / "test_vspectra.pdf",
        plotvspecpol=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        timemin=11,
        timemax=12,
    )

    arr_time_d = np.array(mockplot.call_args_list[0][0][1])
    assert all(np.array_equal(arr_time_d, np.array(mockplot.call_args_list[vspecdir][0][1])) for vspecdir in range(10))

    arr_allvspec = np.vstack([np.array(mockplot.call_args_list[vspecdir][0][2]) for vspecdir in range(10)])

    assert np.allclose(
        arr_allvspec.std(axis=1),
        [
            2.01529689e-12,
            2.05807110e-12,
            2.01551623e-12,
            2.18216916e-12,
            2.85477069e-12,
            3.34384407e-12,
            2.94892344e-12,
            2.29084411e-12,
            2.05916843e-12,
            2.00515984e-12,
        ],
        atol=1e-16,
    )

    assert np.allclose(
        arr_allvspec.mean(axis=1),
        [
            2.9864681492951925e-12,
            3.0063451037690416e-12,
            2.9785924608537284e-12,
            3.2028094816751935e-12,
            4.097482117229833e-12,
            4.663450168092402e-12,
            4.231106733071208e-12,
            3.350080172063692e-12,
            3.0234533505898177e-12,
            2.9721539798925583e-12,
        ],
        atol=1e-16,
    )


@mock.patch.object(mplax.Axes, "plot", side_effect=mplax.Axes.plot, autospec=True)
@pytest.mark.benchmark
def test_vpkt_frompackets_spectrum_plot(mockplot: t.Any, benchmark: BenchmarkFixture) -> None:
    at.spectra.plot(
        argsraw=[],
        specpath=[at.get_config()["path_testdata"] / "vpktcontrib"],
        outputfile=at.get_config()["path_testoutput"] / "test_vpktscontrib_spectra.pdf",
        plotvspecpol=[0, 1, 2, 3, 4, 5, 6, 7, 8],
        frompackets=True,
        maxpacketfiles=2,
        timemin=130,
        timemax=135,
    )

    arr_time_d = np.array(mockplot.call_args_list[0][0][1])
    assert all(np.array_equal(arr_time_d, np.array(mockplot.call_args_list[vspecdir][0][1])) for vspecdir in range(9))

    arr_allvspec = np.vstack([np.array(mockplot.call_args_list[vspecdir][0][2]) for vspecdir in range(9)])
    print(list(arr_allvspec.std(axis=1)))
    assert np.allclose(
        arr_allvspec.std(axis=1),
        [
            2.250169934569213e-15,
            1.5100666673940305e-14,
            4.869648743602329e-15,
            2.2088584226201254e-15,
            1.4772661475549563e-14,
            4.81122573011189e-15,
            2.3113250841019364e-15,
            1.45676358684038e-14,
            4.589891035117792e-15,
        ],
        rtol=0.0001,
    )

    print(list(arr_allvspec.mean(axis=1)))
    assert np.allclose(
        arr_allvspec.mean(axis=1),
        [
            1.1359234428755927e-15,
            9.615508017438282e-15,
            2.959189512289016e-15,
            1.1078180792141745e-15,
            9.453979816323599e-15,
            2.981925804875244e-15,
            1.1359823034815934e-15,
            9.316807974058103e-15,
            2.9236313925953637e-15,
        ],
        rtol=0.0001,
    )



from artistools.packets.packets import add_derived_columns as add_derived_columns
from artistools.packets.packets import add_derived_columns_lazy as add_derived_columns_lazy
from artistools.packets.packets import add_packet_directions_lazypolars as add_packet_directions_lazypolars
from artistools.packets.packets import bin_and_sum as bin_and_sum
from artistools.packets.packets import bin_packet_directions as bin_packet_directions
from artistools.packets.packets import bin_packet_directions_polars as bin_packet_directions_polars
from artistools.packets.packets import get_directionbin as get_directionbin
from artistools.packets.packets import (
    get_mean_packet_emission_velocity_per_ts as get_mean_packet_emission_velocity_per_ts,
)
from artistools.packets.packets import get_packets_batch_parquet_paths as get_packets_batch_parquet_paths
from artistools.packets.packets import get_packets_pl as get_packets_pl
from artistools.packets.packets import get_packets_text_paths as get_packets_text_paths
from artistools.packets.packets import get_virtual_packets_pl as get_virtual_packets_pl
from artistools.packets.packets import make_3d_grid as make_3d_grid
from artistools.packets.packets import make_3d_histogram_from_packets as make_3d_histogram_from_packets
from artistools.packets.packets import readfile as readfile
from artistools.packets.packets import readfile_text as readfile_text
from artistools.packets.packets import type_ids as type_ids



# PYTHON_ARGCOMPLETE_OK
import argparse
import typing as t
from collections.abc import Sequence
from pathlib import Path

import argcomplete
from yaml import dump as yamldump

import artistools as at


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-inputpath", "-i", default=".", help="Path of input ARTIS model")

    parser.add_argument("-temperature", "-T", default=10000, help="Temperature to use in TARDIS file")

    parser.add_argument("-dilution_factor", "-W", default=1.0, help="Dilution factor to use in TARDIS file")

    parser.add_argument(
        "-abundtype",
        choices=["nuclear", "elemental"],
        default="elemental",
        help="Output nuclear or elemental abundances",
    )

    parser.add_argument("-maxatomicnumber", default=92, help="Maximum atomic number for elemental abundances")

    parser.add_argument("-outputpath", "-o", default=".", help="Path of output TARDIS model file")


def main(args: argparse.Namespace | None = None, argsraw: Sequence[str] | None = None, **kwargs: t.Any) -> None:
    """Convert an ARTIS format model to TARDIS format."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)

    temperature = args.temperature
    dilution_factor = args.dilution_factor

    modelpath = Path(args.inputpath)

    pldfmodel, modelmeta = at.inputmodel.get_modeldata(modelpath, get_elemabundances=(args.abundtype == "elemental"))
    t_model_init_days = modelmeta["t_model_init_days"]

    dfmodel = pldfmodel.collect().to_pandas(use_pyarrow_extension_array=True)
    dfmodel.loc[:, "rho"] = 10 ** dfmodel["logrho"]

    if args.abundtype == "nuclear":
        # nuclide abundances
        listspecies = [
            col[2:]
            for col in dfmodel.columns
            if col.startswith("X_") and col.upper() != "X_FEGROUP" and col[-1].isdigit()
        ]
    else:
        # nuclide abundances
        listspecies = [
            col[2:]
            for col in dfmodel.columns
            if col.startswith("X_") and col.upper() != "X_FEGROUP" and not col[-1].isdigit()
        ]

    if args.maxatomicnumber and args.maxatomicnumber > 0:
        listspecies = [species for species in listspecies if at.get_atomic_number(species) <= args.maxatomicnumber]

    modelname = at.get_model_name(modelpath)
    outputfilepath = Path(args.outputpath, f"{modelname}.csvy")
    dictmeta = {
        "name": modelname,
        "description": "This model was converted from ARTIS format with artistools",
        "model_density_time_0": f"{t_model_init_days} day",
        "model_isotope_time_0": f"{t_model_init_days} day",
        "tardis_model_config_version": "v1.0",
        "datatype": {
            "fields": [
                {"name": "velocity", "unit": "km/s", "desc": "velocities of shell outer boundaries"},
                {"name": "density", "unit": "g/cm^3", "desc": "density of shell"},
                {"name": "t_rad", "unit": "K", "desc": "radiative temperature"},
                {"name": "dilution_factor", "desc": "dilution factor of shell"},
                *[{"name": strnuc, "desc": f"fractional {strnuc} abundance"} for strnuc in listspecies],
            ]
        },
    }
    with outputfilepath.open("w", encoding="utf-8") as fileout:
        fileout.write("---\n")
        yamldump(dictmeta, fileout, sort_keys=False)
        fileout.write("---\n")
        fileout.write(",".join(["velocity", "density", "t_rad", "dilution_factor", *listspecies]))
        fileout.write("\n")

        # fileout.write(f'{0.},{0.:.4e},{0.},{0.},{",".join([f"{0.:.4e}" for _ in listspecies])}\n')

        for cell in dfmodel.itertuples(index=False):
            abundlist = [f"{getattr(cell, f'X_{strnuc}'):.4e}" for strnuc in listspecies]
            fileout.write(
                f"{cell.vel_r_max_kmps},{cell.rho:.4e},{temperature},{dilution_factor},{','.join(abundlist)}\n"
            )

    print(f"open {outputfilepath}")


if __name__ == "__main__":
    main()



"""Prepare data for ARTIS KN calculation from end-to-end hydro models. Original script by Oliver Just with modifications by Gerrit Leck for abundance mapping."""

# PYTHON_ARGCOMPLETE_OK
import argparse
import itertools
import typing as t
from collections.abc import Sequence
from pathlib import Path

import argcomplete
import numpy as np
import numpy.typing as npt
import polars as pl

import artistools as at

cl = 29979245800.0
day = 86400.0
msol = 1.989e33  # solar mass in g
tsnap = 0.1 * day  # snapshot time is fixed by the npz files


def sphkernel(
    dist: npt.NDArray[np.floating], hsph: float | npt.NDArray[np.floating], nu: float
) -> npt.NDArray[np.floating]:
    # smoothing kernel for SPH-like interpolation of particle
    # data

    q = dist / hsph
    w = np.where(q < 1.0, 1.0 - 1.5 * q**2 + 0.75 * q**3, np.where(q < 2.0, 0.25 * (2.0 - q) ** 3, 0.0))

    if nu == 3:
        sigma = 1.0 / np.pi
    elif nu == 2:
        sigma = 10.0 / (7.0 * np.pi)

    return w * sigma / hsph**nu


# *******************************************************************


def f1corr(rcyl: npt.NDArray[np.floating], hsph: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
    # correction factor to improve behavior near the axis
    # see Garcia-Senz et al Mon. Not. R. Astron. Soc. 392, 346-360 (2009)

    xi = abs(rcyl) / hsph
    return np.where(
        xi < 1.0,
        1.0 / (7.0 / 15.0 / xi + 2.0 / 3.0 * xi - 1.0 / 6.0 * xi**3 + 1.0 / 20.0 * xi**4),
        np.where(
            xi < 2.0,
            1.0
            / (
                8.0 / 15.0 / xi
                - 1.0 / 3.0
                + 4.0 / 3.0 * xi
                - 2.0 / 3.0 * xi**2
                + 1.0 / 6.0 * xi**3
                - 1.0 / 60.0 * xi**4
            ),
            1.0,
        ),
    )


def get_grid(
    dat_path: Path,
    iso_path: Path,
    vmax: float,
    numb_cells_ARTIS_radial: int,
    numb_cells_ARTIS_z: int,
    nodynej: bool,
    nohmns: bool,
    notorus: bool,
    no_nu_trapping: bool,
    outputpath: Path,
) -> tuple[
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    npt.NDArray[np.floating],
    int,
]:
    dat = np.load(dat_path)
    iso = np.load(iso_path)

    # assume equatorial symmetry? (1=no, 2=yes)
    # determine whether the model assumes equatorial symmetry. Criterion: if the input npz files contains
    # tracer particles only in the upper half space (z > 0 or angle < pi/2), equatorial symmetry is assumed
    # and the reflection w.r.t. the z-axis for the final model.txt has to be done
    eqsymfac = 2 if np.amax(dat.f.pos[:, 1]) < np.pi / 2.0 else 1

    # first re-construct the original post-merger trajectories by merging the
    # splitted dynamical ejecta trajectories
    idx = np.array([round(i) for i in dat.f.idx])  # unique particle ID of all trajectories
    state = np.array([round(i) for i in dat.f.state])  # == -1,0,1 for dynamical, NS-torus, BH-torus
    dyncond = state == -1
    dynidall = np.array([i % 10000 for i in idx])
    dynid = {i % 10000 for i in idx[dyncond]}  # original IDs of dynamical ejecta
    ndyn = len(dynid)
    nodid = {i % 10000 for i in idx[~dyncond]}  # IDs of other ejecta trajs.
    nnod = len(nodid)
    ntraj = ndyn + nnod
    mtraj = np.zeros(ntraj)  # final trajectory mass
    isoA0 = iso[:, 0] + iso[:, 1]  # mass number = neutron number + proton number
    xiso0 = dat.f.nz[:, :] * isoA0[:]  # number fraction -> mass fraction
    ncomp = len(xiso0[0, :])  # number of isotopes
    xtraj = np.zeros((ntraj, ncomp))  # final mass fractions for each isotope at t = tsnap
    vtraj = np.zeros(ntraj)  # final radial velocity
    atraj = np.zeros(ntraj)  # final polar angle
    qtraj = np.zeros(ntraj)  # integrated energy release up to snapshot
    yetraj = np.zeros(ntraj)  # initial electron fraction

    time_s = dat.f.time
    starting_idx = (np.abs(time_s - 0.1)).argmin()
    closest_idx = (np.abs(time_s - tsnap)).argmin()
    i = -1
    tot_Q_rel = 0

    # first get masses and see if they have to be set to zero if the corresponding ejecta types shall be excluded
    # also set to integrated energy release up to snapshot time to zero

    mass_arr = dat.f.mass.copy()
    qdot_arr = dat.f.qdot.copy()

    # ... non-dynamical ejecta
    if nohmns:
        # exclude HMNS ejecta
        mass_arr[np.where(state == 0)] = 1e-30
        qdot_arr[np.where(state == 0)] = 1e-30
    if notorus:
        # exclude torus ejecta
        mass_arr[np.where(state == 1)] = 1e-30
        qdot_arr[np.where(state == 1)] = 1e-30

    hnuloss_arr = dat.f.hnuloss.copy() if no_nu_trapping else np.zeros_like(qdot_arr, dtype=np.float64)

    for i1 in nodid:  # index of Oli's original list
        i += 1  # index in the new list accounting for unprocessed trajs.
        i2 = list(dynidall).index(i1)  # index in Zeweis extended list of trajs.
        mtraj[i] = mass_arr[i2] * msol
        # no multiplication with mass to keep it a specific energy release
        time_by_t_snap = time_s / tsnap
        qtraj[i] = np.trapezoid(
            time_by_t_snap[starting_idx:closest_idx] * (qdot_arr[i2] + hnuloss_arr[i2])[starting_idx:closest_idx],
            time_s[starting_idx:closest_idx],
        )
        tot_Q_rel += mtraj[i] * qtraj[i]

    # ... dynamical ejecta

    if nodynej:
        # exclude dynamical ejecta
        mass_arr[np.where(state == -1)] = 1e-30
        qdot_arr[np.where(state == -1)] = 1e-30
        hnuloss_arr[np.where(state == -1)] = 1e-30
    for i1 in dynid:
        i += 1  # index in the new list accounting for unprocessed trajs.
        i3 = np.where(dynidall == i1)[0]  # indices in Zeweis extended list of trajs.
        mtraj[i] = np.sum(mass_arr[i3]) * msol
        time_by_t_snap = time_s / tsnap
        qtraj[i] = np.trapezoid(
            time_by_t_snap[starting_idx:closest_idx]
            * np.sum((qdot_arr[i3] + hnuloss_arr[i3]), axis=0)[starting_idx:closest_idx],
            time_s[starting_idx:closest_idx],
        )
        tot_Q_rel += mtraj[i] * qtraj[i]

    print(f"tot_Q_rel: {tot_Q_rel}")

    i = -1
    # ... non-dynamical ejecta
    for i1 in nodid:  # index of Oli's original list
        i += 1  # index in the new list accounting for unprocessed trajs.
        i2 = list(dynidall).index(i1)  # index in Zeweis extended list of trajs.
        xtraj[i, :] = xiso0[i2, :]
        # ttraj[i] = dattem.f.T9[i2] * 1e9
        yetraj[i] = dat.f.t5out[i2, 4]
        vtraj[i] = dat.f.pos[i2, 0]
        atraj[i] = dat.f.pos[i2, 1]
    # ... dynamical ejecta
    for i1 in dynid:  # index of my original list
        i += 1  # index in the new list accounting for unprocessed trajs.
        i4 = np.where(dynidall == i1)[0]  # indices in Zeweis extended list of trajs.
        # if len(i2)<nsplit:
        #     print('missing dyn ejecta at i=',i,len(i2))
        weights = dat.f.mass[i4] * msol / (np.sum(dat.f.mass[i4]) * msol)
        xtraj[i, :] = np.sum(weights * xiso0[i4, :].T, axis=1)
        # ttraj[i] = sum(weights * dattem.f.T9[i2] * 1e9)
        # yetraj[i] = np.sum(weights * ye_summ_file[int(i1)])
        yetraj[i] = np.sum(weights * dat.f.t5out[i4, 4])
        vtraj[i] = np.sum(weights * dat.f.pos[i4, 0])
        atraj[i] = np.sum(weights * dat.f.pos[i4, 1])

    # now do the mapping using an SPH like interpolation
    # (see e.g. Price 2007, http://adsabs.harvard.edu/abs/2007PASA...24..159P,
    #  Price & Monaghan 2007, https://ui.adsabs.harvard.edu/abs/2007MNRAS.374.1347P,
    #  and Garcia-Senz 2009, https://ui.adsabs.harvard.edu/abs/2009MNRAS.392..346G)

    # ... smoothing length prefactor and number of dimensions (see Eq. 10 of P2007)
    hsmeta = 1.01
    nu = 2
    # ... cylindrical coordinates of the particle positions
    rcyltraj, zcyltraj = np.zeros(ntraj), np.zeros(ntraj)
    for i in [int(j) for j in np.arange(ntraj)]:
        rcyltraj[i] = vtraj[i] * np.sin(atraj[i]) * cl * tsnap
        zcyltraj[i] = vtraj[i] * np.cos(atraj[i]) * cl * tsnap

    # ... cylindrical coordinates of the grid onto which we want to map
    vmax_cmps = vmax * cl
    nvr = numb_cells_ARTIS_radial
    nvz = numb_cells_ARTIS_z // eqsymfac  # number of mapping grid cells in z direction depends on equatorial symmetry

    wid_init_rcyl = vmax_cmps * tsnap / nvr
    pos_rcyl_min = np.array([vmax_cmps * tsnap / nvr * nr for nr in range(nvr)])
    pos_rcyl_mid = pos_rcyl_min + 0.5 * wid_init_rcyl
    pos_rcyl_max = pos_rcyl_min + wid_init_rcyl

    wid_init_z = (2.0 / eqsymfac) * vmax_cmps * tsnap / nvz
    if eqsymfac == 1:
        pos_z_min = np.array([-vmax_cmps * tsnap + 2.0 * vmax_cmps * tsnap / nvz * nz for nz in range(nvz)])
    else:
        pos_z_min = np.array([vmax_cmps * tsnap / nvz * nz for nz in range(nvz)])
    pos_z_mid = pos_z_min + 0.5 * wid_init_z
    # pos_z_max = pos_z_min + wid_init_z

    rgridc2d = np.array([pos_rcyl_mid[n_r] for n_r in range(nvr) for n_z in range(nvz)]).reshape(nvr, nvz)
    # the z-grid has to be shifted to starting from zero to keep consistency with Oli's script
    zgridc2d = np.array([pos_z_mid[n_z] for n_r in range(nvr) for n_z in range(nvz)]).reshape(nvr, nvz)

    volgrid2d = np.array([
        wid_init_z * np.pi * (pos_rcyl_max[n_r] ** 2 - pos_rcyl_min[n_r] ** 2)
        for n_r in range(nvr)
        for n_z in range(nvz)
    ]).reshape(nvr, nvz)

    # compute mass density and smoothing length of each particle
    # by solving Eq. 10 of P2007 where rho is replaced by the
    # 2D density rho_2D = rho_3D/(2 \pi R) = \sum_i m_i W_2D
    # with particle masses m_i and 2D interpolation kernel W_2D
    print("computing particle densities...")
    rho2dtraj = np.zeros(ntraj)  # this is the 2D density!!!
    hsmooth = np.zeros(ntraj)
    for i in [int(j) for j in np.arange(ntraj)]:
        # print(i)
        cont = True
        hl, hr = 0.00001 * cl * tsnap, 1.0 * cl * tsnap
        dist = np.sqrt((rcyltraj[i] - rcyltraj) ** 2 + (zcyltraj[i] - zcyltraj) ** 2)
        ic = 0
        while cont:
            ic += 1
            h1 = 0.5 * (hl + hr)
            wsph = sphkernel(dist, h1, nu)
            rhos = np.sum(wsph * mtraj)
            fun = (mtraj[i] / ((h1 / hsmeta) ** nu) - rhos) / rhos
            if fun > 0.0:
                hl = h1
            else:
                hr = h1
            if abs(hr - hl) / hl < 1e-5:
                cont = False
                hsmooth[i] = 0.5 * (hl + hr)
                wsph = sphkernel(dist, 0.5 * (hl + hr), nu)
                rho2dtraj[i] = np.sum(wsph * mtraj)
            if ic > 50:
                print("Not good:", ic, hl, hr, fun)
                if ic > 60:
                    msg = "ic > 60"
                    raise AssertionError(msg)

    # f1 correction a la Garcia-Senz? (does not seem to make a significant difference)
    rho2dhat = rho2dtraj * f1corr(rcyltraj, hsmooth)

    # cross check: count number of neighbors within smoothing length
    neinum = np.zeros(ntraj)
    for i in [int(j) for j in np.arange(ntraj)]:
        dist = np.sqrt((rcyltraj[i] - rcyltraj) ** 2 + (zcyltraj[i] - zcyltraj) ** 2)
        neinum[i] = np.sum(np.where(dist / hsmooth < 2.0, 1.0, 0.0))
    neinumavg = np.sum(neinum * mtraj) / np.sum(mtraj)
    print("average number of neighbors:", neinumavg)

    # now interpolate all quantities onto the grid
    print("interpolating...")
    oa = np.add.outer
    distall = np.sqrt(oa(rgridc2d, -rcyltraj) ** 2 + oa(zgridc2d, -zcyltraj) ** 2)
    hall = np.multiply.outer(np.ones((nvr, nvz)), hsmooth)
    wall = sphkernel(distall, hall, nu)
    weight = wall * (mtraj / rho2dhat)
    weinor = (weight.T / (np.sum(weight, axis=2) + 1.0e-100).T).T
    hint = np.sum(weinor * hsmooth, axis=2)
    # ... density
    rho2d = np.sum(wall * mtraj * rho2dtraj / rho2dhat, axis=2)
    rhoint = rho2d / (
        2.0 * np.pi * np.clip(rgridc2d, 0.5 * hint, None)
    )  # limiting to 0.5*h seems to prevent artefacts near the axis
    # ... mass fractions
    xint = np.tensordot(xtraj.T, wall * mtraj, axes=(1, 2)) / (np.sum(wall * mtraj, axis=2) + 1e-100)
    xin2 = np.tensordot(xtraj.T, weinor, axes=(1, 2))  # for testing
    # ... temperature
    qinterpol = np.sum(weinor * qtraj, axis=2)
    yeinterpol = np.sum(weinor * yetraj, axis=2)

    # renormalize so that interpolated mass = sum of particle masses
    dmgrid = rhoint * volgrid2d
    print("total mass after interpolation (but BEFORE renormalization):", np.sum(dmgrid) / msol * eqsymfac)
    rescfac = np.sum(mtraj) / np.sum(dmgrid)
    dmgrid *= rescfac
    rhoint *= rescfac  # the densities are returned by means of rhoint
    mtot = np.sum(dmgrid)

    # test outputs
    print("===> mapped data")
    print("total mass                :", mtot / msol * eqsymfac)
    print("total element mass He Z=2 :", np.sum(np.sum(xint[iso[:, 1] == 2, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass Zr Z=40:", np.sum(np.sum(xint[iso[:, 1] == 40, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass Sn Z=50:", np.sum(np.sum(xint[iso[:, 1] == 50, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass Te Z=52:", np.sum(np.sum(xint[iso[:, 1] == 52, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass Xe Z=54:", np.sum(np.sum(xint[iso[:, 1] == 54, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass W  Z=74:", np.sum(np.sum(xint[iso[:, 1] == 74, :, :], axis=0) * dmgrid) * eqsymfac / msol)
    print("total element mass Pt Z=78:", np.sum(np.sum(xint[iso[:, 1] == 78, :, :], axis=0) * dmgrid) * eqsymfac / msol)

    print("===> tracer data")
    print("total mass                :", np.sum(dat.f.mass) * eqsymfac)
    print("total element mass He Z=2 :", np.sum(np.sum(xiso0[:, iso[:, 1] == 2], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass Zr Z=40:", np.sum(np.sum(xiso0[:, iso[:, 1] == 40], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass Sn Z=50:", np.sum(np.sum(xiso0[:, iso[:, 1] == 50], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass Te Z=52:", np.sum(np.sum(xiso0[:, iso[:, 1] == 52], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass Xe Z=54:", np.sum(np.sum(xiso0[:, iso[:, 1] == 54], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass W  Z=74:", np.sum(np.sum(xiso0[:, iso[:, 1] == 74], axis=1) * dat.f.mass) * eqsymfac)
    print("total element mass Pt Z=78:", np.sum(np.sum(xiso0[:, iso[:, 1] == 78], axis=1) * dat.f.mass) * eqsymfac)

    test = np.sum(xint, axis=0) - 1.0
    test = np.where(test > -1, test, 0.0)
    print("(X-1)_max over 2D grid    :", np.amax(np.where(test > -1, abs(test), 0.0)))

    test = np.sum(xin2, axis=0) - 1.0
    test = np.where(test > -1, test, 0.0)
    print("(X-1)_max over 2D grid 2  :", np.amax(np.where(test > -1, abs(test), 0.0)))

    # write file containing the contribution of each trajectory to each interpolated grid cell
    with (outputpath / Path("gridcontributions.txt")).open("w", encoding="utf-8") as fgridcontributions:
        fgridcontributions.write("particleid cellindex frac_of_cellmass frac_of_cellmass_includemissing" + "\n")
        for nz in np.arange(nvz):
            for nr in np.arange(nvr):
                cellid = nz * nvr + nr + 1
                if dmgrid[nr, nz] > (1e-100 * mtot):
                    # print(
                    # f"{nr} {nz} {temint[nr, nz]} {q_ergperg[nr, nz]} {rhoint[nr, nz]} {dmgrid[nr, nz]} {xint[nr, nz]}"
                    # )
                    wloc = wall[nr, nz, :] * rho2dtraj / rho2dhat
                    wloc /= np.sum(wloc)
                    pids = np.where(wloc > 1.0e-20)[0]
                    for pid in pids:
                        fgridcontributions.write(f"{pid:<10}  {cellid:<8} {wloc[pid]:25.15e} {wloc[pid]:25.15e}\n")

    return rgridc2d, zgridc2d, rhoint, xint, iso, qinterpol, yeinterpol, eqsymfac


def z_reflect(arr: npt.NDArray[np.floating | np.integer], sign: int = 1) -> npt.NDArray[np.floating | np.integer]:
    """Flatten an array and add a reflection in z."""
    _ngridrcyl = arr.shape[0]
    reflected = np.concatenate([sign * np.flip(arr[:, :], axis=1), arr[:, :]], axis=1)
    assert isinstance(reflected, np.ndarray)
    return reflected


# function added by Luke and Gerrit to create the ARTIS model.txt
def create_ARTIS_modelfile(
    ngridrcyl: int,
    ngridz: int,
    vmax: float,
    pos_t_s_grid_rad: npt.NDArray[np.floating | np.integer],
    pos_t_s_grid_z: npt.NDArray[np.floating | np.integer],
    rho_interpol: npt.NDArray[np.floating | np.integer],
    X_cells: npt.NDArray[np.floating | np.integer],
    isot_table: npt.NDArray[t.Any],
    q_ergperg: npt.NDArray[np.floating | np.integer],
    ye_traj: npt.NDArray[np.floating | np.integer],
    eqsymfac: int,
    outputpath: Path,
) -> None:
    numb_cells = ngridrcyl * ngridz
    import pandas as pd

    # now reflect the arrays if equatorial symmetry is assumed or otherwise not
    if eqsymfac == 1:
        dfmodel = pd.DataFrame({
            "inputcellid": range(1, numb_cells + 1),
            "pos_rcyl_mid": (pos_t_s_grid_rad).flatten(order="F"),
            "pos_z_mid": (pos_t_s_grid_z).flatten(order="F"),
            "rho": (rho_interpol).flatten(order="F"),
            "q": (q_ergperg).flatten(order="F"),
            # "cellYe": z_reflect(ye).flatten(order="F"),
        })
    else:
        # equatorial symmetry -> have to reflect
        pos_t_s_grid_rad = z_reflect(pos_t_s_grid_rad)
        pos_t_s_grid_z = z_reflect(pos_t_s_grid_z, sign=-1)
        rho_interpol = z_reflect(rho_interpol)
        q_ergperg = z_reflect(q_ergperg)
        ye_traj = z_reflect(ye_traj)
        dfmodel = pd.DataFrame({
            "inputcellid": range(1, numb_cells + 1),
            "pos_rcyl_mid": (pos_t_s_grid_rad).flatten(order="F"),
            "pos_z_mid": (pos_t_s_grid_z).flatten(order="F"),
            "rho": (rho_interpol).flatten(order="F"),
            "q": (q_ergperg).flatten(order="F"),
            "cellYe": (ye_traj).flatten(order="F"),
        })

    assert pos_t_s_grid_rad.shape == (ngridrcyl, ngridz)
    assert pos_t_s_grid_z.shape == (ngridrcyl, ngridz)
    assert rho_interpol.shape == (ngridrcyl, ngridz)
    assert q_ergperg.shape == (ngridrcyl, ngridz)

    # DF_model, DF_el_contribs, DF_contribs = at.inputmodel.rprocess_from_trajectory.add_abundancecontributions(
    #     dfgridcontributions=DF_gridcontributions,
    #     dfmodel=DF_model,
    #     t_model_days_incpremerger=0.1,
    #     traj_root=Path("./traj_PM/"),
    # )

    # add mass fraction columns
    if "X_Fegroup" not in dfmodel.columns:
        dfmodel = pd.concat([dfmodel, pd.DataFrame({"X_Fegroup": np.ones(len(dfmodel))})], axis=1)
    # pdb.set_trace()

    dictabunds = {}
    dictelabunds = {"inputcellid": np.array(range(1, numb_cells + 1))}
    for tuple_idx, isot_tuple in enumerate(isot_table):
        if eqsymfac == 1:
            flat_isoabund = np.nan_to_num((X_cells[tuple_idx]).flatten(order="F"), nan=0.0)
        else:
            flat_isoabund = np.nan_to_num(z_reflect(X_cells[tuple_idx]).flatten(order="F"), nan=0.0)
        if np.any(flat_isoabund):
            elem_str = f"X_{at.get_elsymbol(isot_tuple[1])}"
            isotope_str = f"{elem_str}{isot_tuple[0] + isot_tuple[1]}"
            dictabunds[isotope_str] = flat_isoabund
            dictelabunds[elem_str] = (
                dictelabunds[elem_str] + flat_isoabund if elem_str in dictelabunds else flat_isoabund
            )

    print(f"Number of non-zero nuclides {len(dictabunds)}")
    dfmodel = pd.concat([dfmodel, pd.DataFrame(dictabunds)], axis=1)

    dfabundances = pd.DataFrame(dictelabunds).fillna(0.0)

    # create init abundance file
    at.inputmodel.save_initelemabundances(dfelabundances=pl.from_pandas(dfabundances), outpath=outputpath)

    # create modelmeta dictionary
    modelmeta = {
        "dimensions": 2,
        "ncoordgridrcyl": ngridrcyl,
        "ncoordgridz": ngridz,
        "t_model_init_days": tsnap / day,
        "vmax_cmps": vmax * cl,
    }

    # create model.txt
    at.inputmodel.save_modeldata(dfmodel=pl.from_pandas(dfmodel), modelmeta=modelmeta, outpath=outputpath)


def get_old_cell_indices(red_fact: int, new_r: int, new_z: int, N_cell_r_old: int) -> list[int]:
    # function to get old grid indices for a given new grid index
    old_r_indices = np.arange(red_fact * (new_r - 1) + 1, red_fact * new_r + 1)
    old_z_indices = np.arange(red_fact * (new_z - 1), red_fact * new_z) * N_cell_r_old
    indices = np.add.outer(old_r_indices, old_z_indices).flatten().tolist()
    return sorted(int(x) for x in indices)


def remap_mass_weighted_quantity(
    im_ARTIS_old: pl.DataFrame,
    fieldname: str,
    red_fact: int,
    N_cell_r_new: int,
    N_cell_z_new: int,
    N_cell_r_old: int,
    Delta_r: float,
    Delta_z: float,
) -> npt.NDArray[np.float64]:
    # function that remaps any mass-weighted quantity
    new_numb_cells = N_cell_r_new * N_cell_z_new
    new_field = np.zeros(new_numb_cells)

    # get arrays
    mass_arr = im_ARTIS_old["mass_g"].to_numpy()
    field_arr = im_ARTIS_old[fieldname].to_numpy()

    for new_z, new_r in itertools.product(range(1, N_cell_z_new + 1), range(1, N_cell_r_new + 1)):
        new_cell_idx = new_r + N_cell_r_new * (new_z - 1)

        old_idxs = get_old_cell_indices(red_fact, new_r, new_z, N_cell_r_old)
        old_idxs_np = np.array(old_idxs) - 1

        masses = mass_arr[old_idxs_np]
        values = field_arr[old_idxs_np]

        if fieldname == "mass_g":
            r_i = (new_r - 1) * Delta_r
            r_o = new_r * Delta_r
            V_new = np.pi * (r_o**2 - r_i**2) * Delta_z
            new_field[new_cell_idx - 1] = masses.sum() / V_new
        else:
            new_field[new_cell_idx - 1] = np.average(values, weights=masses)

    return new_field


def merge_ARTIS_cells(red_fact: int, N_r: int, N_z: int, v_max: float, outputpath: Path) -> None:
    """red_fact: number of cells to be merged."""
    import pandas as pd

    red_fact_1D = int(np.sqrt(red_fact))
    im_ARTIS_old = at.inputmodel.get_modeldata(modelpath=Path(), derived_cols=["mass_g"])[0].collect()
    N_cell_r_old, N_cell_z_old = N_r, N_z
    N_cell_r_new, N_cell_z_new = int(N_cell_r_old / red_fact_1D), int(N_cell_z_old / red_fact_1D)
    new_numb_cells = N_cell_r_new * N_cell_z_new
    r_max_snap = v_max * cl * tsnap

    # create new grid
    Delta_r = r_max_snap / N_cell_r_new
    Delta_z = 2 * r_max_snap / N_cell_z_new  # account for positive and negative z-values
    assert np.isclose(Delta_r, Delta_z, rtol=1e-20), "New grid cells not quadratic"
    r_mid_grid = np.array([(i + 0.5) * Delta_r for i in range(N_cell_r_new)])
    z_mid_grid = np.array([-r_max_snap + (i + 0.5) * Delta_z for i in range(N_cell_z_new)])

    print("Now remap masses, q and Ye")
    # remap density, integrated energy release and Y_e
    new_rho_arr = remap_mass_weighted_quantity(
        im_ARTIS_old, "mass_g", red_fact_1D, N_cell_r_new, N_cell_z_new, N_r, Delta_r, Delta_z
    )
    new_q_arr = remap_mass_weighted_quantity(
        im_ARTIS_old, "q", red_fact_1D, N_cell_r_new, N_cell_z_new, N_r, Delta_r, Delta_z
    )
    new_ye_arr = remap_mass_weighted_quantity(
        im_ARTIS_old, "Ye", red_fact_1D, N_cell_r_new, N_cell_z_new, N_r, Delta_r, Delta_z
    )
    print("  Done.")

    # new model data frame
    dfmodel = pd.DataFrame({
        "inputcellid": range(1, new_numb_cells + 1),
        "pos_rcyl_mid": np.tile(r_mid_grid, N_cell_z_new),
        "pos_z_mid": np.repeat(z_mid_grid, N_cell_r_new),
        "rho": new_rho_arr,
        "q": new_q_arr,
        "cellYe": new_ye_arr,
    })

    # now map the abundances
    dictabunds = {}
    element_abbrevs_list = [at.get_elsymbol(Z) for Z in range(1, 101)]  # Keep full list as in your code
    element_abbrevs_list_titled = [abbrev.title() for abbrev in element_abbrevs_list]
    el_mass_fracs = np.zeros((len(element_abbrevs_list), new_numb_cells))
    dictelabunds = {"inputcellid": np.array(range(1, new_numb_cells + 1))}

    nuclide_columns = [col for col in im_ARTIS_old.columns if col.startswith("X_")][1:]

    masses_all = im_ARTIS_old["mass_g"].to_numpy()
    abunds_all = {nuclide: im_ARTIS_old[nuclide].to_numpy() for nuclide in nuclide_columns}
    for nuclide in nuclide_columns:
        new_X_arr = np.zeros(new_numb_cells)
        for new_z in range(1, N_cell_z_new + 1):
            for new_r in range(1, N_cell_r_new + 1):
                new_cell_idx = new_r + N_cell_r_new * (new_z - 1)
                old_idxs = get_old_cell_indices(red_fact_1D, new_r, new_z, N_cell_r_old)
                old_idxs_np = np.array(old_idxs) - 1
                masses = masses_all[old_idxs_np]
                abunds = abunds_all[nuclide][old_idxs_np]
                new_X_arr[new_cell_idx - 1] = np.average(abunds, weights=masses)
        dictabunds[nuclide] = new_X_arr
        el = "".join([i for i in nuclide[2:] if not i.isdigit()])
        if el in element_abbrevs_list_titled:
            el_idx = element_abbrevs_list_titled.index(el)
            el_mass_fracs[el_idx] += new_X_arr

    for i, el in enumerate(element_abbrevs_list_titled):
        dictelabunds[f"X_{el}"] = el_mass_fracs[i]

    if "X_Fegroup" not in dfmodel.columns:
        dfmodel = pd.concat([dfmodel, pd.DataFrame({"X_Fegroup": np.ones(len(dfmodel))})], axis=1)

    dfmodel = pd.concat([dfmodel, pd.DataFrame(dictabunds)], axis=1)
    dfabundances = pd.DataFrame(dictelabunds).fillna(0.0)

    at.inputmodel.save_initelemabundances(dfelabundances=pl.from_pandas(dfabundances), outpath=outputpath)
    modelmeta = {
        "dimensions": 2,
        "ncoordgridrcyl": N_cell_r_new,
        "ncoordgridz": N_cell_z_new,
        "t_model_init_days": tsnap / day,
        "vmax_cmps": v_max * cl,
    }

    at.inputmodel.save_modeldata(dfmodel=pl.from_pandas(dfmodel), modelmeta=modelmeta, outpath=outputpath)
    print(f"Successfully remapped model to {N_cell_r_new}x{N_cell_z_new} grid.")


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-outputpath", "-o", default=".", help="Path of output ARTIS model file")

    parser.add_argument("-npz", required=True, help="Path to the model npz file")

    parser.add_argument(
        "-iso",
        default=None,
        help="Path to the nuclide information npy file. If not provided, will be assumed to be [npz_path]/iso_table.npy",
    )

    parser.add_argument(
        "-vmax", type=float, default=0.5, help="Maximum one-direction velocity in units of c the ARTIS model shall have"
    )

    parser.add_argument(
        "-ngridrcyl", type=int, default=25, help="Number of cells in radial direction the ARTIS model shall have"
    )

    parser.add_argument(
        "-ngridz", type=int, default=50, help="Number of cells in z direction the ARTIS model shall have"
    )

    parser.add_argument("--nonutrapping", action="store_true", help="Exclude neutrino energy from the snapshot energy")

    parser.add_argument("--nodyn", action="store_true", help="Exclude dynamical ejecta from the model")

    parser.add_argument("--nohmns", action="store_true", help="Exclude neutrino wind ejecta from the model")

    parser.add_argument("--notorus", action="store_true", help="Exclude torus ejecta from the model")

    parser.add_argument(
        "-mergecells",
        type=int,
        default=None,
        help="Merge specified number of cells in postprocessing to keep precision in the mass fractions",
    )


def main(args: argparse.Namespace | None = None, argsraw: Sequence[str] | None = None, **kwargs: t.Any) -> None:
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)

        addargs(parser)
        at.set_args_from_dict(parser, kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args([] if kwargs else argsraw)
    if args.iso is None:
        args.iso = Path(args.npz).parent / "iso_table.npy"

    numb_cells_ARTIS_radial = int(args.ngridrcyl)
    numb_cells_ARTIS_z = int(args.ngridz)
    pos_t_s_grid_rad, pos_t_s_grid_z, rho_interpol, X_cells, isot_table, q_ergperg, ye_traj, eqsymfac = get_grid(
        args.npz,
        args.iso,
        float(args.vmax),
        numb_cells_ARTIS_radial,
        numb_cells_ARTIS_z,
        args.nodyn,
        args.nohmns,
        args.notorus,
        no_nu_trapping=args.nonutrapping,
        outputpath=args.outputpath,
    )

    create_ARTIS_modelfile(
        numb_cells_ARTIS_radial,
        numb_cells_ARTIS_z,
        float(args.vmax),
        pos_t_s_grid_rad,
        pos_t_s_grid_z,
        rho_interpol,
        X_cells,
        isot_table,
        q_ergperg,
        ye_traj,
        eqsymfac,
        args.outputpath,
    )

    if args.mergecells is not None:
        # bunch of checks to assure proper cell merging
        assert isinstance(args.mergecells, int), "Number of cells to merge is not an integer!"
        # check if the number is a square number
        assert np.sqrt(args.mergecells).is_integer(), "Number of cells to merge is not a square number!"
        # check if the current number of cells is a multiple of the cells to merge
        assert (numb_cells_ARTIS_z / np.sqrt(args.mergecells)).is_integer(), (
            "Number of merged cells in z direction is no integer!"
        )
        assert (numb_cells_ARTIS_radial / np.sqrt(args.mergecells)).is_integer(), (
            "Number of merged cells in r direction is no integer!"
        )
        merge_ARTIS_cells(
            args.mergecells, numb_cells_ARTIS_radial, numb_cells_ARTIS_z, float(args.vmax), args.outputpath
        )


if __name__ == "__main__":
    main()



from pathlib import Path


def main() -> None:
    directions_costheta_phi = [(1, 0), (0, 0), (-1, 0)]

    # Nspectra opacity choices (i.e. Nspectra spectra for each observer)
    # 0: full opacity, -1: no line opacity; -2: no bf opacity; -3: no ff opacity; -4: no es opacity,
    # +ve: exclude element with atomic number's contribution to bound-bound opacity
    opacityexclusions: list[int] = []

    # time window. If override_tminmax=1 it restrict vpkt to time windown
    override_tminmax = 0
    vspec_tmin_in_days = 0.2
    vspec_tmax_in_days = 1.5

    custom_lambda_ranges: list[tuple[float, float]] = [(3500, 18000)]

    override_thickcell_tau = True  # if override_thickcell_tau=1  vpkt are not created when cell optical depth is larger than cell_is_optically_thick_vpkt
    cell_is_optically_thick_vpkt = 100

    # maximum optical depth before vpkt is thrown away
    tau_max_vpkt = 10

    # produce velocity grid map?
    vgrid_on = False

    # Specify time range for velocity grid map. Used if vgrid_on=1
    tmin_vgrid_in_days = 0.2
    tmax_vgrid_in_days = 1.5

    # Specify wavelength range for velocity grid map: number of intervals (Nrange_grid) and limits (dum10,dum11)
    Nrange_grid = 1
    vgrid_lambda_min = 3500
    vgrid_lambda_max = 6000  # can have multiple ranges -- not implemented

    str_custom_lambda_ranges = (
        (f" {len(custom_lambda_ranges)}" + " ".join(f"{lmin} {lmax}" for lmin, lmax in custom_lambda_ranges))
        if custom_lambda_ranges
        else ""
    )

    new_vpktfile = Path() / "vpkt.txt"
    with new_vpktfile.open("w", encoding="utf-8") as vpktfile:
        vpktfile.write(
            f"{len(directions_costheta_phi)}\n"
            f"{' '.join(str(costheta) for costheta, _ in directions_costheta_phi)}\n"
            f"{' '.join(str(phi) for _, phi in directions_costheta_phi)}\n"
            f"{bool(opacityexclusions):d} {f'{len(opacityexclusions)} ' + ' '.join(str(x) for x in opacityexclusions) if opacityexclusions else ''}\n"
            f"{override_tminmax} {vspec_tmin_in_days} {vspec_tmax_in_days}\n"
            f"{bool(custom_lambda_ranges):d}{str_custom_lambda_ranges}\n"
            f"{override_thickcell_tau:d} {cell_is_optically_thick_vpkt}\n"
            f"{tau_max_vpkt}\n"
            f"{vgrid_on:d}\n"
            f"{tmin_vgrid_in_days} {tmax_vgrid_in_days}\n"
            f"{Nrange_grid} {vgrid_lambda_min} {vgrid_lambda_max}"  # this can have multiple wavelength ranges. May need changed.
        )


if __name__ == "__main__":
    main()
